import ast
import builtins
import copy
import importlib
import inspect
import traceback
import types
import warnings
from contextlib import contextmanager

from funcparserlib.parser import NoParseError, many

import hy
from hy.compat import PY3_14
from hy.errors import HyCompileError, HyLanguageError, HySyntaxError
from hy.macros import macroexpand
from hy.model_patterns import FORM, KEYWORD, unpack
from hy.models import (
    Bytes,
    Complex,
    Dict,
    Expression,
    FComponent,
    Float,
    FString,
    Integer,
    Keyword,
    Lazy,
    List,
    Object,
    Set,
    String,
    Symbol,
    Tuple,
    as_model,
    is_unpack,
)
from hy.reader import mangle, HyReader
from hy.scoping import ResolveOuterVars, ScopeGlobal


def calling_module(n=1):
    """Get the module calling, if available.

    As a fallback, this will import a module using the calling frame's
    globals value of `__name__`.

    Args:
        n (int): The number of levels up the stack from this function call.
            The default is `1` (level up).

    Returns:
        types.ModuleType: The module at stack level `n + 1` or `None`.
    """
    frame_up = inspect.stack(0)[n + 1][0]
    module = inspect.getmodule(frame_up)
    if module is None:
        # This works for modules like `__main__`
        module_name = frame_up.f_globals.get("__name__", None)
        if module_name:
            try:
                module = importlib.import_module(module_name)
            except ImportError:
                pass
    return module


_model_compilers = {}


def builds_model(*model_types):
    "Assign the decorated method to _model_compilers for the given types."

    def _dec(fn):
        for t in model_types:
            _model_compilers[t] = fn
        return fn

    return _dec


# Provide asty.Foo(x, ...) as shorthand for
# ast.Foo(..., lineno=x.start_line, col_offset=x.start_column) or
# ast.Foo(..., lineno=x.lineno, col_offset=x.col_offset)
# Also provides asty.parse(x, ...) which recursively
# copies x's position data onto the parse result.
class Asty:
    POS_ATTRS = {
        "lineno": "start_line",
        "col_offset": "start_column",
        "end_lineno": "end_line",
        "end_col_offset": "end_column",
    }

    @staticmethod
    def _get_pos(node):
        return {
            attr: getattr(node, hy_attr, getattr(node, attr, None))
            for attr, hy_attr in Asty.POS_ATTRS.items()
        }

    @staticmethod
    def _replace_pos(node, pos):
        for attr, value in pos.items():
            if hasattr(node, attr):
                setattr(node, attr, value)
        for child in ast.iter_child_nodes(node):
            Asty._replace_pos(child, pos)

    def parse(self, x, *args, **kwargs):
        res = ast.parse(*args, **kwargs)
        Asty._replace_pos(res, Asty._get_pos(x))
        return res

    def __getattr__(self, name):
        setattr(
            Asty,
            name,
            staticmethod(
                lambda x, **kwargs: getattr(ast, name)(**Asty._get_pos(x), **kwargs)
            ),
        )
        return getattr(Asty, name)


asty = Asty()


class Result:
    """
    Smart representation of the result of a hy->AST compilation

    This object tries to reconcile the hy world, where everything can be used
    as an expression, with the Python world, where statements and expressions
    need to coexist.

    To do so, we represent a compiler result as a list of statements `stmts`,
    terminated by an expression context `expr`. The expression context is used
    when the compiler needs to use the result as an expression.

    Results are chained by addition: adding two results together returns a
    Result representing the succession of the two Results' statements, with
    the second Result's expression context.

    We make sure that a non-empty expression context does not get clobbered by
    adding more results, by checking accesses to the expression context. We
    assume that the context has been used, or deliberately ignored, if it has
    been accessed.

    The Result object is interoperable with python AST objects: when an AST
    object gets added to a Result object, it gets converted on-the-fly.
    """

    __slots__ = ("stmts", "temp_variables", "_expr", "__used_expr")

    def __init__(self, *, stmts=(), expr=None, temp_variables=()):
        self.stmts = list(stmts)
        self.temp_variables = list(temp_variables)
        self._expr = expr

        self.__used_expr = False

    @property
    def expr(self):
        self.__used_expr = True
        return self._expr

    @expr.setter
    def expr(self, value):
        self.__used_expr = False
        self._expr = value

    @property
    def lineno(self):
        if self._expr is not None:
            return self._expr.lineno
        if self.stmts:
            return self.stmts[-1].lineno
        return None

    @property
    def col_offset(self):
        if self._expr is not None:
            return self._expr.col_offset
        if self.stmts:
            return self.stmts[-1].col_offset
        return None

    @property
    def end_col_offset(self):
        if self._expr is not None:
            return self._expr.end_col_offset
        if self.stmts:
            return self.stmts[-1].end_col_offset
        return None

    @property
    def end_lineno(self):
        if self._expr is not None:
            return self._expr.end_lineno
        if self.stmts:
            return self.stmts[-1].end_lineno
        return None

    def is_expr(self):
        """Check whether I am a pure expression"""
        return self._expr and not self.stmts

    @property
    def force_expr(self):
        """Force the expression context of the Result.

        If there is no expression context, we return a "None" expression.
        """
        if self.expr:
            return self.expr
        return ast.Constant(
            value=None,
            lineno=self.stmts[-1].lineno if self.stmts else 0,
            col_offset=self.stmts[-1].col_offset if self.stmts else 0,
        )

    def expr_as_stmt(self):
        """Convert the Result's expression context to a statement

        This is useful when we want to use the stored expression in a
        statement context (for instance in a code branch).

        We drop ast.Names if they are appended to statements, as they
        can't have any side effect. "Bare" names still get converted to
        statements.

        If there is no expression context, return an empty result.
        """
        if self.expr and not (isinstance(self.expr, ast.Name) and self.stmts):
            return Result() + asty.Expr(self.expr, value=self.expr)
        return Result()

    def rename(self, compiler, new_name):
        """Rename the Result's temporary variables to a `new_name`.

        We know how to handle ast.Names and ast.FunctionDefs.
        """
        new_name = mangle(new_name)
        for var in self.temp_variables:
            if isinstance(var, ast.Name):
                var.id = new_name
                compiler.scope.assign(var)
            elif isinstance(var, (ast.FunctionDef, ast.AsyncFunctionDef)):
                var.name = new_name
            else:
                raise TypeError(
                    "Don't know how to rename a %s!" % (var.__class__.__name__)
                )
        self.temp_variables = []

    def __add__(self, other):
        # If we add an ast statement, convert it first
        if isinstance(other, ast.stmt):
            return self + Result(stmts=[other])

        # If we add an ast expression, clobber the expression context
        if isinstance(other, ast.expr):
            return self + Result(expr=other)

        if isinstance(other, ast.excepthandler):
            return self + Result(stmts=[other])

        if not isinstance(other, Result):
            raise TypeError(f"Can't add {self!r} with non-compiler result {other!r}")

        # Check for expression context clobbering
        if self.expr and not self.__used_expr:
            traceback.print_stack()
            print(
                "Bad boy clobbered expr {} with {}".format(
                    ast.dump(self.expr), ast.dump(other.expr)
                )
            )

        # Fairly obvious addition
        result = Result()
        result.stmts = self.stmts + other.stmts
        result.expr = other.expr
        result.temp_variables = other.temp_variables

        return result

    def __str__(self):
        return "Result(stmts=[%s], expr=%s)" % (
            ", ".join(ast.dump(x) for x in self.stmts),
            ast.dump(self.expr) if self.expr else None,
        )


def make_hy_model(outer, x, rest):
    return outer(
        [Symbol(a) if type(a) is str else a[0] if type(a) is list else a for a in x]
        + (rest or [])
    )


def mkexpr(*items, **kwargs):
    return make_hy_model(Expression, items, kwargs.get("rest"))


def is_annotate_expression(model):
    return isinstance(model, Expression) and model and model[0] == Symbol("annotate")


class HyASTCompiler:
    """A Hy-to-Python AST compiler"""

    def __init__(self, module, filename=None, source=None, extra_macros=None):
        """
        Args:
            module (Union[str, types.ModuleType]): Module name or object in which the Hy tree is evaluated.
            filename (Optional[str]): The name of the file for the source to be compiled.
                This is optional information for informative error messages and
                debugging.
            source (Optional[str]): The source for the file, if any, being compiled.  This is optional
                information for informative error messages and debugging.
            extra_macros (Optional[dict]): More macros to use during lookup. They take precedence
                over macros in `module`.
        """
        self.anon_var_count = 0
        self.temp_if = None
        self.extra_macros = extra_macros or {}

        # Make a list of dictionaries with local compiler settings,
        # such as the definitions of local macros. The last element is
        # considered the top of the stack.
        self.local_state_stack = []
        self.new_local_state()

        if not inspect.ismodule(module):
            self.module = importlib.import_module(module)
        else:
            self.module = module
        self.module.hy = hy
        # The `hy` module itself should always be in scope.

        self.module_name = self.module.__name__

        self.filename = filename
        self.source = source

        self.this = None
        # Set in `macroexpand` to the current expression being
        # macro-expanded, so it can be accessed as `&compiler.this`.

        # Hy expects this to be present, so we prep the module for Hy
        # compilation.
        self.module.__dict__.setdefault("_hy_macros", {})
        self.module.__dict__.setdefault("_hy_reader_macros", {})

        self.scope = ScopeGlobal(self)

    def new_local_state(self):
        'Add a new local state to the top of the stack.'
        self.local_state_stack.append(dict(macros = {}))

    def is_in_local_state(self):
        return len(self.local_state_stack) > 1

    def get_local_option(self, key, default):
        'Get the topmost available value of a local-state setting.'
        return next(
            (s[key]
                for s in reversed(self.local_state_stack)
                if key in s),
            default)

    def warn_on_core_shadow(self, name):
        if (
                mangle(name) in getattr(builtins, "_hy_macros", {}) and
                self.get_local_option('warn_on_core_shadow', True)):
            warnings.warn(
                f"New macro `{name}` will shadow the core macro of the same name",
                RuntimeWarning
            )

    def get_anon_var(self, base="anon", name=""):
        self.anon_var_count += 1
        name = name and "_" + name
        return f"_hy_{base}{name}_{self.anon_var_count}"

    def compile_atom(self, atom):
        # Compilation methods may mutate the atom, so copy it first.
        atom = copy.copy(atom)
        return Result() + _model_compilers[type(atom)](self, atom)

    def compile(self, tree):
        if tree is None:
            return Result()
        try:
            ret = self.compile_atom(tree)
            return ret
        except HyCompileError:
            # compile calls compile, so we're going to have multiple raise
            # nested; so let's re-raise this exception, let's not wrap it in
            # another HyCompileError!
            raise
        except HyLanguageError as e:
            # These are expected errors that should be passed to the user.
            raise e
        except Exception as e:
            # These are unexpected errors that will--hopefully--never be seen
            # by the user.
            f_exc = traceback.format_exc()
            exc_msg = "Internal Compiler Bug\n {}".format(f_exc)
            raise HyCompileError(exc_msg)

    def _syntax_error(self, expr, message):
        raise HySyntaxError(message, expr, self.filename, self.source)

    def _compile_collect(self, exprs, with_kwargs=False, dict_display=False):
        """Collect the expression contexts from a list of compiled expression.

        This returns a list of the expression contexts, and the sum of the
        Result objects passed as arguments.

        """
        compiled_exprs = []
        ret = Result()
        keywords = []

        exprs_iter = iter(exprs)
        for expr in exprs_iter:

            if is_unpack("mapping", expr):
                ret += self.compile(expr[1])
                if dict_display:
                    compiled_exprs.append(None)
                    compiled_exprs.append(ret.force_expr)
                elif with_kwargs:
                    keywords.append(asty.keyword(expr, arg=None, value=ret.force_expr))

            elif with_kwargs and isinstance(expr, Keyword):
                try:
                    value = next(exprs_iter)
                except StopIteration:
                    raise self._syntax_error(
                        expr, "Keyword argument {kw} needs a value.".format(kw=expr)
                    )

                if not expr:
                    raise self._syntax_error(
                        expr, "Can't call a function with the empty keyword"
                    )

                compiled_value = self.compile(value)
                ret += compiled_value

                arg = str(expr)[1:]
                keywords.append(
                    asty.keyword(expr, arg=mangle(arg), value=compiled_value.force_expr)
                )

            else:
                ret += self.compile(expr)
                compiled_exprs.append(ret.force_expr)

        return compiled_exprs, ret, keywords

    @builds_model(Lazy)
    def _compile_branch(self, exprs):
        """Make a branch out of an iterable of Result objects

        This generates a Result from the given sequence of Results, forcing each
        expression context as a statement before the next result is used.

        We keep the expression context of the last argument for the returned Result
        """
        result = Result()
        last = None
        for node in exprs:
            if last is not None:
                result += last.expr_as_stmt()
            last = self.compile(node)
            result += last
        return result

    def _storeize(self, expr, name, func=None):
        """Return a new `name` object with an ast.Store() context"""
        if not func:
            func = ast.Store

        if isinstance(name, Result):
            if not name.is_expr():
                raise self._syntax_error(
                    expr, "Can't assign or delete a non-expression"
                )
            name = name.expr

        if isinstance(name, (ast.Tuple, ast.List)):
            typ = type(name)
            new_elts = []
            for x in name.elts:
                new_elts.append(self._storeize(expr, x, func))
            new_name = typ(elts=new_elts)
        elif isinstance(name, ast.Name):
            new_name = ast.Name(id=name.id)
            if func == ast.Store:
                self.scope.assign(new_name)
        elif isinstance(name, ast.Subscript):
            new_name = ast.Subscript(value=name.value, slice=name.slice)
        elif isinstance(name, ast.Attribute):
            new_name = ast.Attribute(value=name.value, attr=name.attr)
        elif isinstance(name, ast.Starred):
            new_name = ast.Starred(value=self._storeize(expr, name.value, func))
        else:
            raise self._syntax_error(
                expr,
                "Can't assign or delete a "
                + (
                    "constant"
                    if isinstance(name, ast.Constant)
                    else type(expr).__name__
                ),
            )

        new_name.ctx = func()
        ast.copy_location(new_name, name)
        return new_name

    def _nonconst(self, name):
        if str(name) in ("None", "True", "False"):
            raise self._syntax_error(name, "Can't assign to constant")
        return name

    def eval(self, model):
        return hy_eval(
            model,
            locals = self.module.__dict__,
            module = self.module,
            filename = self.filename,
            source = self.source,
            import_stdlib = False)

    @contextmanager
    def local_state(self):
        self.new_local_state()
        try:
            yield
        finally:
            self.local_state_stack.pop()

    @builds_model(Expression)
    def compile_expression(self, expr):
        # Perform macro expansions
        expr = macroexpand(expr, self.module, self)
        if isinstance(expr, (Result, ast.AST)):
            # Use this as-is.
            return expr
        elif not isinstance(expr, Expression):
            # Go through compile again if we have a different type of model.
            return self.compile(expr)

        if not expr:
            raise self._syntax_error(
                expr, "empty expressions are not allowed at top level"
            )

        args = list(expr)
        root = args.pop(0)
        func = None

        if (
            isinstance(root, Expression)
            and len(root) >= 2
            and isinstance(root[0], Symbol)
            and not str(root[0]).strip(".")
            and root[1] == Symbol("None")
        ):
            # ((. None a1 a2) obj v1 v2) -> ((. obj a1 a2) v1 v2)
            # (The reader already parsed `.a1.a2` as `(. None a1 a2)`.)

            # Find the object we're calling the method on.
            i = 0
            while i < len(args):
                if isinstance(args[i], Keyword):
                    if i == 0 and len(args) == 1:
                        break
                    i += 2
                elif is_unpack("iterable", args[i]):
                    raise self._syntax_error(
                        args[i], "can't call a method on an `unpack-iterable` form"
                    )
                elif is_unpack("mapping", args[i]):
                    i += 1
                else:
                    break
            else:
                raise self._syntax_error(expr, "attribute access requires object")

            func = self.compile(
                Expression([Symbol("."), args.pop(i), *root[2:]]).replace(root)
            )

        if is_annotate_expression(root):
            # Flatten and compile the annotation expression.
            ann_expr = Expression(root + args).replace(root)
            return self.compile_expression(ann_expr)

        if not func:
            func = self.compile(root)

        args, ret, keywords = self._compile_collect(args, with_kwargs=True)

        return (
            func + ret + asty.Call(expr, func=func.expr, args=args, keywords=keywords)
        )

    @builds_model(Integer, Float, Complex)
    def compile_numeric_literal(self, x):
        return asty.Constant(x, value =
            {Integer: int, Float: float, Complex: complex}[type(x)](x))

    @builds_model(Symbol)
    def compile_symbol(self, symbol):
        if symbol == Symbol("..."):
            return asty.Constant(symbol, value=Ellipsis)

        # By this point, `symbol` should be either all dots or
        # dot-free.
        assert not symbol.strip(".") or "." not in symbol

        if mangle(symbol) in ("None", "False", "True"):
            return asty.Constant(symbol, value=ast.literal_eval(mangle(symbol)))

        return self.scope.access(asty.Name(symbol, id=mangle(symbol), ctx=ast.Load()))

    @builds_model(Keyword)
    def compile_keyword(self, obj):
        ret = Result()
        ret += asty.Call(
            obj,
            func=asty.Attribute(
                obj,
                value=asty.Attribute(
                    obj,
                    value=asty.Name(obj, id="hy", ctx=ast.Load()),
                    attr="models",
                    ctx=ast.Load(),
                ),
                attr="Keyword",
                ctx=ast.Load(),
            ),
            args=[asty.Constant(obj, value=obj.name)],
            keywords=[],
        )
        return ret

    @builds_model(String, Bytes)
    def compile_string(self, string):
        return asty.Constant(string, value =
            (bytes if type(string) is Bytes else str)(string))

    @builds_model(FComponent)
    def compile_fcomponent(self, fcomponent):
        if not PY3_14 and fcomponent.is_tstring:
            raise self._syntax_error(fcomponent, "Template string literals require Python 3.14 or later")
        if fcomponent.conversion not in (None, 's', 'r', 'a'):
            raise self._syntax_error(
                fcomponent, f"Invalid conversion character {fcomponent.conversion!r}"
            )
        conversion = ord(fcomponent.conversion) if fcomponent.conversion else -1
        root, *rest = fcomponent
        value = self.compile(root)
        elts, ret, _ = self._compile_collect(rest)
        if elts:
            spec = asty.JoinedStr(fcomponent, values=elts)
        else:
            spec = None
        return (
            value
            + ret
            + (asty.Interpolation if fcomponent.is_tstring else asty.FormattedValue)(
                fcomponent, value=value.expr, conversion=conversion, format_spec=spec,
                **(dict(str=fcomponent.expression) if fcomponent.is_tstring else {}),
            )
        )

    @builds_model(FString)
    def compile_fstring(self, fstring):
        if not PY3_14 and fstring.is_tstring:
            raise self._syntax_error(fstring, "Template string literals require Python 3.14 or later")
        elts, ret, _ = self._compile_collect(fstring)
        return ret + (asty.TemplateStr if fstring.is_tstring else asty.JoinedStr)(fstring, values=elts)

    @builds_model(List, Set)
    def compile_list(self, expression):
        elts, ret, _ = self._compile_collect(expression)
        node = {List: asty.List, Set: asty.Set}[type(expression)]
        return ret + node(
            expression,
            elts = elts,
            **({} if node is asty.Set else dict(ctx = ast.Load())))

    @builds_model(Dict)
    def compile_dict(self, m):
        keyvalues, ret, _ = self._compile_collect(m, dict_display=True)
        return ret + asty.Dict(m, keys=keyvalues[::2], values=keyvalues[1::2])

    @builds_model(Tuple)
    def compile_tuple(self, expression):
        elts, ret, _ = self._compile_collect(expression)
        return ret + asty.Tuple(expression, elts=elts, ctx=ast.Load())


def get_compiler_module(module=None, compiler=None, calling_frame=False):
    """Get a module object from a compiler, given module object,
    string name of a module, and (optionally) the calling frame; otherwise,
    raise an error."""

    module = getattr(compiler, "module", None) or module

    if isinstance(module, str):
        module = importlib.import_module(mangle(module))

    if calling_frame and not module:
        module = calling_module(n=2)

    if not inspect.ismodule(module):
        raise TypeError("Invalid module type: {}".format(type(module)))

    return module


def hy_eval(
    hytree,
    locals,
    module=None,
    compiler=None,
    filename=None,
    source=None,
    import_stdlib=True,
    globals=None,
    extra_macros=None,
):

    module = get_compiler_module(module, compiler, True)

    # Does the Hy AST object come with its own information?
    filename = getattr(hytree, "filename", filename) or "<string>"
    source = getattr(hytree, "source", source)

    _ast, expr = hy_compile(
        hytree,
        module,
        get_expr=True,
        compiler=compiler,
        filename=filename,
        source=source,
        import_stdlib=import_stdlib,
        extra_macros=extra_macros,
    )

    if globals is None:
        globals = module.__dict__

    # Two-step eval: eval() the body of the exec call
    eval(compile(_ast, filename, "exec"), globals, locals)

    # Then eval the expression context and return that
    return eval(compile(expr, filename, "eval"), globals, locals)


def hy_eval_user(model, globals = None, locals = None, module = None, macros = None):
    # This function is advertised as `hy.eval`.
    """An equivalent of Python's :func:`eval` for evaluating Hy code. The chief difference is that the first argument should be a :ref:`model <models>` rather than source text. If you have a string of source text you want to evaluate, convert it to a model first with :hy:func:`hy.read` or :hy:func:`hy.read-many`::

        (hy.eval '(+ 1 1))             ; => 2
        (hy.eval (hy.read "(+ 1 1)"))  ; => 2

    The optional arguments ``globals`` and ``locals`` work as in the case of :func:`eval`.

    Another optional argument, ``module``, can be a module object or a string naming a module. The module's ``__dict__`` attribute can fill in for ``globals`` (and hence also for ``locals``) if ``module`` is provided but ``globals`` isn't, but the primary purpose of ``module`` is to control where macro calls are looked up. Without this argument, the calling module of ``hy.eval`` is used instead. ::

        (defmacro my-test-mac [] 3)
        (hy.eval '(my-test-mac))                 ; => 3
        (import hyrule)
        (hy.eval '(my-test-mac) :module hyrule)  ; NameError
        (hy.eval '(list-n 3 1) :module hyrule)   ; => [1 1 1]

    Finally, finer control of macro lookup can be achieved by passing in a dictionary of macros as the ``macros`` argument. The keys of this dictionary should be mangled macro names, and the values should be function objects to implement those macros. This is the same structure as is produced by :hy:func:`local-macros <hy.core.macros.local-macros>`, and in fact, ``(hy.eval … :macros (local-macros))`` is useful to make local macros visible to ``hy.eval``, which otherwise doesn't see them. ::

        (defn f []
          (defmacro lmac [] 1)
          (hy.eval '(lmac))     ; NameError
          (print (hy.eval '(lmac) :macros (local-macros)))) ; => 1
        (f)

    In any case, macros provided in this dictionary will shadow macros of the same name that are associated with the provided or implicit module. You can shadow a core macro, too, so be careful: there's no warning for this as there is in the case of :hy:func:`defmacro`."""

    if locals is None:
        locals = globals
    hy_was = None
    if locals and 'hy' in locals:
        hy_was = (locals['hy'],)
    try:
        value = hy_eval(
            hytree = model,
            globals = globals,
            locals = (inspect.getargvalues(inspect.stack()[1][0]).locals
                if locals is None and module is None
                else locals),
            module = get_compiler_module(module, None, True),
            extra_macros = macros)
    finally:
        if locals is not None:
            if hy_was:
                # Restore the old value of `hy`.
                locals['hy'], = hy_was
            else:
                # Remove the implicitly added `hy` (if execution
                # reached far enough to add it).
                locals.pop('hy', None)
    return value


def hy_compile(
    tree,
    module,
    root=ast.Module,
    get_expr=False,
    compiler=None,
    filename=None,
    source=None,
    import_stdlib=True,
    extra_macros=None,
):
    """Compile a hy.models.Object tree into a Python AST Module.

    Args:
        tree (Object): The Hy AST object to compile.
        module (Union[str, types.ModuleType]): Module, or name of the module, in which the Hy tree is evaluated.
            The module associated with `compiler` takes priority over this value.
        root (Type[ast.AST]): Root object for the Python AST tree.
        get_expr (bool): If true, return a tuple with `(root_obj, last_expression)`.
        compiler (Optional[HyASTCompiler]): An existing Hy compiler to use for compilation.  Also serves as
            the `module` value when given.
        filename (Optional[str]): The filename corresponding to the source for `tree`.  This will be
            overridden by the `filename` field of `tree`, if any; otherwise, it
            defaults to "<string>".  When `compiler` is given, its `filename` field
            value is always used.
        source (Optional[str]): A string containing the source code for `tree`.  This will be
            overridden by the `source` field of `tree`, if any; otherwise,
            if `None`, an attempt will be made to obtain it from the module given by
            `module`.  When `compiler` is given, its `source` field value is always
            used.
        extra_macros (Optional[dict]): Passed through to `HyASTCompiler`, if it's called.

    Returns:
        ast.AST: A Python AST tree
    """
    module = get_compiler_module(module, compiler, False)

    filename = getattr(tree, "filename", filename)
    source = getattr(tree, "source", source)
    reader = getattr(tree, "reader", None)

    tree = as_model(tree)
    if not isinstance(tree, Object):
        raise TypeError(
            "`tree` must be a hy.models.Object or capable of " "being promoted to one"
        )

    compiler = compiler or HyASTCompiler(
        module,
        filename = filename,
        source = source,
        extra_macros = extra_macros)

    with HyReader.using_reader(reader, create=False), compiler.scope:
        result = compiler.compile(tree)
    expr = result.force_expr

    if not get_expr:
        result += result.expr_as_stmt()

    result.stmts = list(map(ResolveOuterVars().visit, result.stmts))

    body = []

    if issubclass(root, ast.Module):
        # Pull out a single docstring and prepend to the resulting body.
        if (
            result.stmts
            and isinstance(result.stmts[0], ast.Expr)
            and isinstance(result.stmts[0].value, ast.Constant)
            and isinstance(result.stmts[0].value.value, str)
        ):

            body += [result.stmts.pop(0)]

        # Pull out any __future__ imports, since they are required to be at the beginning.
        while (
            result.stmts
            and isinstance(result.stmts[0], ast.ImportFrom)
            and result.stmts[0].module == "__future__"
        ):

            body += [result.stmts.pop(0)]

        # Import hy for runtime.
        if import_stdlib:
            body.append(ast.fix_missing_locations(ast.Import([ast.alias("hy", None)])))

    body += result.stmts
    ret = root(
        body = body,
        **({} if root is ast.Interactive else dict(type_ignores = [])))

    if get_expr:
        expr = ast.Expression(body=expr)
        ret = (ret, expr)

    return ret



"""
Get useful information from live Hy or Python objects.

This module provides Hy compatibility with CPython's inspect module.
Within Hy, Hy macros are supported via Hy's `get_macro` macro,
for example `(inspect.getsource (get-macro get-macro))`.
The Hy-specific functionality is monkey-patched into `inspect` at run-time.

The class finder in `findsource` relies on python 3.13 or later.
"""

import inspect
import linecache
from contextlib import suppress

from hy.compat import PY3_13
from hy.errors import HySyntaxError
from hy.models import Expression, Lazy
from hy.reader import HyReader, read
from hy.reader.exceptions import LexException


class HySafeReader(HyReader):
    """A HyReader that skips over non-default reader macros.

    Skip over undefined reader macros so that no code is executed at read time.
    Do this by replacing the tag reader method to return None, as comments do.
    """

    @reader_for(")")
    @reader_for("]")
    @reader_for("}")
    def _skip(self, key):
        return None

    @reader_for("#")
    def _safe_tag_dispatch(self, key):
        """Skip over undefined reader macros so that no code is executed at read time."""
        with suppress(LexException):
            return super().tag_dispatch(key)


def isExpression(object):
    """Check if object is a Hy Expression instance."""
    return isinstance(object, Expression)


def isLazy(object):
    """Check if object is a Hy Lazy instance."""
    return isinstance(object, Lazy)


def getfile(object):
    """Work out which source or compiled file an object was defined in."""

    if isLazy(object) or isExpression(object):
        if getattr(object, "filename", None) and object.filename != "<string>":
            return object.filename
        else:
            # python's inspect.getfile raises OSError where there's no __file__
            raise OSError("source code not available")

    try:
        return py_getfile(object)
    except TypeError:
        # python's inspect.getfile raises TypeError for unhandled types
        raise TypeError(
            "module, class, method, function, traceback, frame, "
            "code, Expression or Lazy object was expected, got "
            "{}".format(type(object).__name__)
        )


def findsource(object):
    """Return the entire source file and starting line number for an object.

    First looks for Hy source, otherwise defers to the original
    `inspect.findsource`. The argument may be a module, class, method,
    function, traceback, frame, code, Lazy or Expression object. The source
    code is returned as a list of all the lines in the file and the line number
    indexes a line in that list. An OSError is raised if the source code cannot
    be retrieved.
    """
    if getfile(object) is None:
        raise OSError("source code not available")

    fname = inspect.getsourcefile(object)

    if isExpression(object) or isLazy(object):
        if getattr(object, "start_line", None) is None:
            raise OSError("source code not available")
        return (linecache.getlines(fname), object.start_line)
    elif getfile(object).endswith(".hy") and not inspect.isframe(object):
        # We identify other Hy objects from the file extension.
        module = inspect.getmodule(object, fname)
        # List of source code lines.
        lines = (
            linecache.getlines(fname, module.__dict__)
            if module
            else linecache.getlines(fname)
        )
        if inspect.ismodule(object):
            return (lines, 0)
        # Some objects already have the information we need.
        elif hasattr(object, "__code__") and hasattr(object.__code__, "co_firstlineno"):
            # This indexing offset is actually correct.
            return (lines, object.__code__.co_firstlineno - 1)
        elif inspect.iscode(object):
            return (lines, object.co_firstlineno - 1)
        elif inspect.isclass(object):
            # _ClassFinder exists and can be used prior to python 3.13,
            # but would require compiling the ast which may be unsafe,
            # so just decline to do it.
            if PY3_13:
                return py_findsource(object)
            else:
                raise OSError("finding Hy class source code requires Python 3.13")
        else:
            raise OSError("source code not available")
    else:
        # Non-Hy object
        return py_findsource(object)


def getcomments(object):
    """Get comments immediately preceding an object's source code.

    First checks for Hy source, otherwise defers to the original `inspect.getcomments`.
    Returns None when the source can't be found.
    """
    try:
        lines, lnum = findsource(object)
    except (OSError, TypeError):
        # Return None when the source can't be found.
        return None

    if getfile(object).endswith(".hy"):
        # Roughly follows the logic of inspect.getcomments, but for Hy comments
        if not lines:
            return None

        comments = []
        if inspect.ismodule(object) or isExpression(object) or isLazy(object):
            # Remove shebang.
            start = 1 if lines and lines[0][:2] == "#!" else 0
            # Remove preceding empty lines and textless comments.
            while start < len(lines) and set(lines[start].strip()) == {";"}:
                start += 1
            if start < len(lines) and lines[start].lstrip().startswith(";"):
                end = start
                while end < len(lines) and lines[end].lstrip().startswith(";"):
                    comments.append(lines[end].expandtabs())
                    end += 1
                return "".join(comments)
            else:
                return None
        # Look for a comment block preceding the object
        elif lnum > 0 and lnum < len(lines):
            # Look for comments above the object and work up.
            end = lnum - 1
            while end >= 0 and lines[end].lstrip().startswith(";"):
                comments = [lines[end].expandtabs().lstrip()] + comments
                end -= 1
            return "".join(comments) if comments else None
        else:
            return None

    else:
        # Non-Hy object
        return py_getcomments(object)


def getsource(object):
    """Return the text of the source code for an object.

    The argument may be a module, class, method, function, traceback, frame,
    or code object.  The source code is returned as a single string.  An
    OSError is raised if the source code cannot be retrieved."""
    return "".join(getsourcelines(object)[0])


def hy_getblock(lines):
    """Extract the lines of code corresponding to the first Hy form from the given list of lines.

    Built-in Hy reader macros are allowed as safe, since they do not execute code.
    User-defined reader macros could execute code, so we use `HySafeReader`.
    """
    # Read the first form and use its attributes
    try:
        form = read("".join(lines), reader=HySafeReader())
    except HySyntaxError as e:
        raise e from None
    return lines[: form.end_line]


def getsourcelines(object):
    """Return a list of source lines and starting line number for a Hy or python object.

    First checks for Hy source, otherwise defers to the original `inspect.getsourcelines`.

    The argument may be a module, class, method, function, traceback, frame,
    code, Lazy or Expression object. The source code is returned as a list of
    the lines corresponding to the object and the line number indicates where
    in the original source file the first line of code was found. An OSError is
    raised if the source code cannot be retrieved.
    """
    object = inspect.unwrap(object)
    lines, lnum = findsource(object)

    if inspect.istraceback(object):
        object = object.tb_frame

    # For module or frame that corresponds to module, return all source lines.
    if inspect.ismodule(object) or (
        inspect.isframe(object) and object.f_code.co_name == "<module>"
    ):
        return lines, 0

    # Almost everything already works with python's inspect.
    # The inspect.getblock function relies on inspect.BlockFinder which
    # assumes python tokenization.
    # So deal with this as a special case using hy_getblock.
    elif getfile(object).endswith(".hy"):
        return hy_getblock(lines[lnum:]), lnum + 1
    else:
        # Non-Hy object
        return py_getsourcelines(object)


if hasattr(inspect, "_hy_originals"):
    # Retrieve saved versions of `inspect`'s original functions.
    py_findsource, py_getcomments, py_getfile, py_getsource, py_getsourcelines = inspect._hy_originals
else:
    # Save the originals and then monkey-patch.
    inspect._hy_originals = \
        py_findsource,       py_getcomments,      py_getfile,      py_getsource,      py_getsourcelines = \
        inspect.findsource,  inspect.getcomments, inspect.getfile, inspect.getsource, inspect.getsourcelines
    inspect.findsource, inspect.getcomments, inspect.getfile, inspect.getsource, inspect.getsourcelines = \
            findsource,         getcomments,         getfile,         getsource,         getsourcelines



"Character reader for parsing Hy source."

import codecs
import inspect
from contextlib import contextmanager, nullcontext
from itertools import islice

import hy
from hy.models import (
    Bytes,
    Complex,
    Dict,
    Expression,
    FComponent,
    Float,
    FString,
    Integer,
    Keyword,
    List,
    Set,
    String,
    Symbol,
    Tuple,
    as_model,
)

from .exceptions import LexException, PrematureEndOfInput
from .reader import Reader, isnormalizedspace


def sym(name):
    return Symbol(name, from_parser=True)


# Note: This is subtly different from
# the `mkexpr` in hy/compiler.py !
def mkexpr(root, *args):
    return Expression((sym(root) if isinstance(root, str) else root, *args))


def as_identifier(ident, reader=None):
    """Generate a Hy model from an identifier.

    Also verifies the syntax of dot notation and validity of symbol names.

    Parameters
    ----------
    ident : str
        Text to convert.

    reader : Reader, optional
        The reader to use, if any; used for generating position data for errors.

    Returns
    -------
    out : a hy.models.Object subtype corresponding to the parsed text.
    """
    try:
        return Integer(ident)
    except ValueError:
        pass
    try:
        return Float(ident)
    except ValueError:
        pass
    if ident not in ("j", "J"):
        try:
            return Complex(ident)
        except ValueError:
            pass

    if "." in ident:
        if not ident.strip("."):
            # It's all dots. Return it as a symbol.
            return sym(ident)

        def err(msg):
            raise (
                ValueError(msg)
                if reader is None
                else LexException.from_reader(msg, reader)
            )

        if ident.lstrip(".").find("..") > 0:
            err(
                "In a dotted identifier, multiple dots in a row are only allowed at the start"
            )
        if ident.endswith("."):
            err("A dotted identifier can't end with a dot")
        head = "." * (len(ident) - len(ident.lstrip(".")))
        args = [as_identifier(a, reader=reader) for a in ident.lstrip(".").split(".")]
        if any(not isinstance(a, Symbol) for a in args):
            err("The parts of a dotted identifier must be symbols")
        return (
            mkexpr(sym("."), *args)
            if head == ""
            else mkexpr(head, Symbol("None"), *args)
        )

    if reader is None:
        if (
            not ident
            or ident[0] in ":#"
            or any(isnormalizedspace(c) for c in ident)
            or HyReader.NON_IDENT.intersection(ident)
        ):
            raise ValueError(f"Syntactically illegal symbol: {ident!r}")

    return sym(ident)


class HyReader(Reader):
    """A modular reader for Hy source. It inherits from
    :py:class:`hy.Reader`.

    When ``use_current_readers`` is true, initialize this reader
    with all reader macros from the calling module."""

    __module__ = 'hy'

    ###
    # Components necessary for Reader implementation
    ###

    NON_IDENT = set("()[]{};\"'`~")
    _current_reader = None

    def __init__(self, *, use_current_readers=False, bracketed_templates=False):
        super().__init__()

        self.bracketed_templates = bracketed_templates

        # move any reader macros declared using
        # `reader_for("#...")` to the macro table
        self.reader_macros = {}
        for tag in list(self.reader_table.keys()):
            if tag[0] == '#' and tag[1:]:
                self.reader_macros[tag[1:]] = self.reader_table.pop(tag)

        if use_current_readers:
            self.reader_macros.update(
                inspect.stack()[1].frame.f_globals.get("_hy_reader_macros", {})
            )

    @classmethod
    def current_reader(cls, override=None, create=True):
        return override or HyReader._current_reader or (cls() if create else None)

    @contextmanager
    def as_current_reader(self):
        old_reader = HyReader._current_reader
        HyReader._current_reader = self
        try:
            yield
        finally:
            HyReader._current_reader = old_reader

    @classmethod
    @contextmanager
    def using_reader(cls, override=None, create=True):
        reader = cls.current_reader(override, create)
        with reader.as_current_reader() if reader else nullcontext():
            yield


    def fill_pos(self, model, start):
        """Set position information for ``model``. ``start`` should be a (line
        number, column number) tuple for the start position, whereas the end
        position is set to the current cursor position."""

        model.start_line, model.start_column = start
        model.end_line, model.end_column = self.pos
        return model.replace(model)
          # `replace` will recurse into submodels and set any model
          # positions that are still unset the same way.

    def read_default(self, key):
        """Try to read an identifier. If the next character after that is
        ``"``, then instead parse it as a string with the given prefix (e.g.,
        ``r"..."``).

        (This method is the default reader handler, for when nothing in the
        read table matches.)"""

        ident = key + self.read_ident()
        if self.peek_and_getc('"'):
            return self.prefixed_string('"', ident)
        return as_identifier(ident, reader=self)

    def parse(self, stream, filename=None, skip_shebang=False):
        """Yield all models in ``stream``. The parameters are understood as in
        :hy:func:`hy.read-many`."""

        self._set_source(stream, filename)

        if skip_shebang and "".join(
                islice(self.peeking(eof_ok = True), len("#!"))) == "#!":
            for c in self.chars():
                if c == "\n":
                    break

        yield from self.parse_forms_until("")

    ###
    # Reading forms
    ###

    def try_parse_one_form(self):
        """Attempt to parse a single Hy form.

        Read one (non-space) character from the stream, then call the
        corresponding handler.

        Returns:
            hy.models.Object | None:
                Model optionally returned by the called handler. Handlers may
                return `None` to signify no parsed form (e.g., for comments).

        Raises:
            PrematureEndOfInput: If the reader hits the end of the file before
                fully parsing a form.
            LexException: If there is an error during form parsing.
        """
        with self.as_current_reader():
            try:
                self.slurp_space()
                c = self.getc()
                start = self._pos
                if not c:
                    raise PrematureEndOfInput.from_reader(
                        "Premature end of input while attempting to parse one form", self
                    )
                handler = self.reader_table.get(c)
                model = handler(self, c) if handler else self.read_default(c)
                if model is not None:
                    model = self.fill_pos(model, start)
                    model.reader = self
                    return model
                return None
            except LexException:
                raise
            except Exception as e:
                raise LexException.from_reader(
                    str(e) or "Exception thrown attempting to parse one form", self
                )

    def parse_one_form(self):
        """Parse the next form in the stream and return its model. Any
        preceding whitespace and comments are skipped over."""
        model = None
        while model is None:
            model = self.try_parse_one_form()
        return model

    def parse_forms_until(self, closer):
        """Yield models until the character ``closer`` is seen. This method is
        useful for reading sequential constructs such as lists."""
        while True:
            self.slurp_space()
            if self.peek_and_getc(closer):
                break
            model = self.try_parse_one_form()
            if model is not None:
                yield model

    ###
    # Basic atoms
    ###

    @reader_for(")")
    @reader_for("]")
    @reader_for("}")
    def INVALID(self, key):
        raise LexException.from_reader(
            f"Ran into a '{key}' where it wasn't expected.", self
        )

    @reader_for(";")
    def line_comment(self, _):
        any(c == "\n" for c in self.chars(eof_ok=True))
        return None

    @reader_for(":")
    def keyword(self, _):
        ident = self.read_ident()
        if "." in ident:
            raise LexException.from_reader(
                "Cannot access attribute on anything other"
                " than a name (in order to get attributes of expressions,"
                " use `(. <expression> <attr>)` or `(.<attr> <expression>)`)",
                self,
            )
        return Keyword(ident, from_parser=True)

    @reader_for('"')
    def prefixed_string(self, _, prefix=""):
        prefix_chars = set(prefix)
        if (
            len(prefix_chars) != len(prefix)
            or not prefix_chars < set("bfrt")
            or len(prefix_chars - set("r")) > 1
        ):
            raise LexException.from_reader(f"invalid string prefix {prefix!r}", self)

        escaping = False

        def quote_closing(c):
            nonlocal escaping
            if c == "\\":
                escaping = not escaping
                return 0
            if c == '"' and not escaping:
                return 1
            if (
                escaping
                and "r" not in prefix
                and
                # https://docs.python.org/3/reference/lexical_analysis.html#string-and-bytes-literals
                c
                not in ("\n\r\\'\"abfnrtv01234567x" + ("" if "b" in prefix else "NuU"))
            ):
                raise LexException.from_reader("invalid escape sequence \\" + c, self)
            escaping = False
            return 0

        return self.read_string_until(quote_closing, prefix, fstring_mode = (
            "f" if "f" in prefix_chars
            else "t" if "t" in prefix_chars
            else ""
        ))

    ###
    # Special annotations
    ###

    @reader_for("'", ("quote",))
    @reader_for("`", ("quasiquote",))
    def tag_as(root):
        return lambda self, _: mkexpr(root, self.parse_one_form())

    @reader_for("~")
    def unquote(self, key):
        return mkexpr(
            "unquote" + ("-splice" if self.peek_and_getc("@") else ""),
            self.parse_one_form(),
        )

    ###
    # Sequences
    ###

    @reader_for("(", (Expression, ")"))
    @reader_for("[", (List, "]"))
    @reader_for("{", (Dict, "}"))
    @reader_for("#{", (Set, "}"))
    @reader_for("#(", (Tuple, ")"))
    def sequence(seq_type, closer):
        return lambda self, _: seq_type(self.parse_forms_until(closer))

    ###
    # Reader tag-macros
    ###

    @reader_for("#")
    def tag_dispatch(self, key):
        """General handler for reader macros (and tag macros).

        Reads a full identifier after the `#` and calls the corresponding handler
        (this allows, e.g., `#reads-multiple-forms foo bar baz`).
        """

        if not self.peekc().strip():
            raise PrematureEndOfInput.from_reader(
                "Premature end of input while attempting dispatch", self
            )

        # try dispatching tagged ident
        ident = self.read_ident() or self.getc()
        if ident in self.reader_macros:
            tree = self.reader_macros[ident](self, ident)
            return as_model(tree) if tree is not None else None

        raise LexException.from_reader(
            f"reader macro '{key + ident}' is not defined", self
        )

    @reader_for("#_")
    def discard(self, _):
        """Discards the next parsed form."""
        self.parse_one_form()
        return None

    @reader_for("#*")
    @reader_for("#**")
    def hash_star(self, stars):
        """Unpacking forms `#*` and `#**`, corresponding to `*` and `**` in Python."""
        return mkexpr(
            "unpack-" + {"*": "iterable", "**": "mapping"}[stars],
            self.parse_one_form(),
        )

    @reader_for("#^")
    def annotate(self, _):
        """Annotate a symbol, usually with a type."""
        typ = self.parse_one_form()
        target = self.parse_one_form()
        return mkexpr("annotate", target, typ)

    ###
    # Strings
    # (these are more complicated because f-strings
    #  form their own sublanguage)
    ###

    @reader_for("#[")
    def bracketed_string(self, _):
        """Bracketed strings. See the Hy docs for full details."""
        delim = []
        for c in self.chars():
            if c == "[":
                break
            elif c == "]":
                raise LexException.from_reader(
                    "Ran into a ']' where it wasn't expected.", self
                )
            delim.append(c)
        delim = "".join(delim)
        fstring_mode = (
            "f" if delim == "f" or delim.startswith("f-")
            else "t" if self.bracketed_templates and (delim == "t" or delim.startswith("t-"))
            else ""
        )

        # discard single initial newline, if any, accounting for all
        # three styles of newline
        self.peek_and_getc("\x0d")
        self.peek_and_getc("\x0a")

        index = -1

        def delim_closing(c):
            nonlocal index
            if c == "]":
                if index == len(delim):
                    # this is the second bracket at the end of the delim
                    return len(delim) + 2
                else:
                    # reset state, this may be the first bracket of closing delim
                    index = 0
            elif 0 <= index <= len(delim):
                # we're inside a possible closing delim
                if index < len(delim) and c == delim[index]:
                    index += 1
                else:
                    # failed delim, reset state
                    index = -1
            return 0

        return self.read_string_until(delim_closing, "r", fstring_mode, brackets=delim)

    def read_string_until(self, closing, prefix, fstring_mode, **kwargs):
        if fstring_mode:
            components = self.read_fcomponents_until(closing, prefix, fstring_mode)
            return FString(components, is_tstring=fstring_mode == 't', **kwargs)
        s = self.read_chars_until(closing, prefix, fstring_mode="")
        return (Bytes if isinstance(s, bytes) else String)(s, **kwargs)

    def read_chars_until(self, closing, prefix, fstring_mode):
        s = []
        in_named_escape = False
        for c in self.chars():
            s.append(c)
            # check if c is closing
            n_closing_chars = closing(c)
            if n_closing_chars:
                # string has ended
                s = s[:-n_closing_chars]
                break
            if fstring_mode:
                # handle braces in f-strings
                if c == "{":
                    if "r" not in prefix and s[-3:] == ["\\", "N", "{"]:
                        # ignore "\N{...}"
                        in_named_escape = True
                    elif not self.peek_and_getc("{"):
                        # start f-component if not "{{"
                        s.pop()
                        break
                elif c == "}":
                    if in_named_escape:
                        in_named_escape = False
                    elif not self.peek_and_getc("}"):
                        raise SyntaxError(f"{fstring_mode}-string: single '}}' is not allowed")
        res = "".join(s).replace("\x0d\x0a", "\x0a").replace("\x0d", "\x0a")

        if "b" in prefix:
            try:
                res = res.encode('ascii')
            except UnicodeEncodeError:
                raise SyntaxError("bytes can only contain ASCII literal characters")

        if "r" not in prefix:
            # perform string escapes
            if "b" in prefix:
                res = codecs.escape_decode(res)[0]
            else:
                # formula taken from https://stackoverflow.com/a/57192592
                # encode first to ISO-8859-1 ("Latin 1") due to a Python bug,
                # see https://github.com/python/cpython/issues/65530
                res = res.encode('ISO-8859-1', errors='backslashreplace').decode('unicode_escape')

        if fstring_mode:
            return res, n_closing_chars
        return res

    def read_fcomponents_until(self, closing, prefix, fstring_mode):
        components = []
        start = self.pos
        while True:
            s, closed = self.read_chars_until(closing, prefix, fstring_mode=fstring_mode)
            if s:
                components.append(self.fill_pos(String(s), start))
            if closed:
                break
            components.extend(self.read_fcomponent(prefix, fstring_mode))
        return components

    def read_fcomponent(self, prefix, fstring_mode):
        """May return one or two components, since the `=` debugging syntax
        will create a String component."""
        start = self.pos
        values = []
        conversion = None
        has_debug = False

        # read the expression, saving the text verbatim
        # in case we encounter debug `=`
        space_before = self.slurp_space()
        with self.saving_chars() as form_text:
            model = self.parse_one_form()
        space_between = self.slurp_space()
        form_text = "".join(form_text)

        # check for and handle debug syntax:
        # we emt the verbatim text before we emit the value
        if self.peek_and_getc("="):
            has_debug = True
            space_after = self.slurp_space()
            dbg_prefix = (
                space_before + form_text + space_between + "=" + space_after
            )
            values.append(self.fill_pos(String(dbg_prefix), start))

        # handle conversion code
        if self.peek_and_getc("!"):
            conversion = self.getc()
        self.slurp_space()

        def component_closing(c):
            if c == "}":
                return 1
            return 0

        # handle formatting options
        format_components = []
        if self.peek_and_getc(":"):
            format_components = self.read_fcomponents_until(component_closing, prefix, "f")
        else:
            if has_debug and conversion is None:
                conversion = "r"
            if not self.getc() == "}":
                raise LexException.from_reader(f"{fstring_mode}-string: trailing junk in field", self)
        return values + [
            self.fill_pos(FComponent((model, *format_components), conversion=conversion, expression=form_text, is_tstring=fstring_mode == "t"), start)
        ]



import re
from io import StringIO

import hy.models

from .hy_reader import HyReader
from .mangling import mangle, unmangle

__all__ = ["mangle", "unmangle", "read", "read_many"]


def read_many(stream, filename="<string>", reader=None, skip_shebang=False):
    """Parse all the Hy source code in ``stream``, which should be a textual file-like
    object or a string. ``filename``, if provided, is used in error messages. If no
    ``reader`` is provided, a new :class:`hy.HyReader` object is created. If
    ``skip_shebang`` is true and a :ref:`shebang line <shebang>` is present, it's
    detected and discarded first.

    Return a value of type :class:`hy.models.Lazy`. If you want to evaluate this, be
    careful to allow evaluating each model before reading the next, as in ``(hy.eval
    (hy.read-many o))``. By contrast, forcing all the code to be read before evaluating
    any of it, as in ``(hy.eval `(do [~@(hy.read-many o)]))``, will yield the wrong
    result if one form defines a reader macro that's later used in the same stream to
    produce new forms.

    .. warning::
       Thanks to reader macros, reading can execute arbitrary code. Don't read untrusted
       input."""

    if isinstance(stream, str):
        stream = StringIO(stream)
    pos = stream.tell()
    source = stream.read()
    stream.seek(pos)

    reader = reader or HyReader()
    m = hy.models.Lazy(reader.parse(
        stream, filename, skip_shebang))
    m.source = source
    m.filename = filename
    m.reader = reader
    return m


def read(stream, filename=None, reader=None):
    """Like :hy:func:`hy.read-many`, but only one form is read, and shebangs are
    forbidden. The model corresponding to this specific form is returned, or, if there
    are no forms left in the stream, :class:`EOFError` is raised. ``stream.pos`` is left
    where it was immediately after the form."""

    it = read_many(stream, filename, reader)
    try:
        m = next(it)
    except StopIteration:
        raise EOFError()
    else:
        m.source, m.filename, m.reader = it.source, it.filename, it.reader
        return m



import builtins
import importlib
import inspect
import os
import pkgutil
import sys
import types
import zipimport
from contextlib import contextmanager
from functools import partial

import hy
from hy.compiler import hy_compile
from hy.reader import read_many, HyReader


@contextmanager
def loader_module_obj(loader):
    """Use the module object associated with a loader.

    This is intended to be used by a loader object itself, and primarily as a
    work-around for attempts to get module and/or file code from a loader
    without actually creating a module object.  Since Hy currently needs the
    module object for macro importing, expansion, and whatnot, using this will
    reconcile Hy with such attempts.

    For example, if we're first compiling a Hy script starting from
    `runpy.run_path`, the Hy compiler will need a valid module object in which
    to run, but, given the way `runpy.run_path` works, there might not be one
    yet (e.g. `__main__` for a .hy file).  We compensate by properly loading
    the module here.

    The function `inspect.getmodule` has a hidden-ish feature that returns
    modules using their associated filenames (via `inspect.modulesbyfile`),
    and, since the Loaders (and their delegate Loaders) carry a filename/path
    associated with the parent package, we use it as a more robust attempt to
    obtain an existing module object.

    When no module object is found, a temporary, minimally sufficient module
    object is created for the duration of the `with` body.
    """
    tmp_mod = False

    try:
        module = inspect.getmodule(None, _filename=loader.path)
    except KeyError:
        module = None

    if module is None:
        tmp_mod = True
        module = sys.modules.setdefault(loader.name, types.ModuleType(loader.name))
        module.__file__ = loader.path
        module.__name__ = loader.name

    try:
        yield module
    finally:
        if tmp_mod:
            del sys.modules[loader.name]


def _hy_code_from_file(filename, loader_type=None):
    """Use PEP-302 loader to produce code for a given Hy source file."""
    full_fname = os.path.abspath(filename)
    fname_path, fname_file = os.path.split(full_fname)
    modname = os.path.splitext(fname_file)[0]
    sys.path.insert(0, fname_path)
    try:
        if loader_type is None:
            loader = pkgutil.get_loader(modname)
        else:
            loader = loader_type(modname, full_fname)
        code = loader.get_code(modname)
    finally:
        sys.path.pop(0)

    return code


def _get_code_from_file(*args, hy_src_check=lambda x: x.endswith(".hy")):
    """A patch of `runpy._get_code_from_file` that will also run and cache Hy
    code.
    """

    if hy.compat.PY3_15:
        fname, module = args
    elif hy.compat.PY3_12:
        fname, = args
    else:
        run_name, fname = args
        if fname is None and run_name is not None:
            fname = run_name

    # Check for bytecode first.  (This is what the `runpy` version does!)
    with open(fname, "rb") as f:
        code = pkgutil.read_code(f)

    if code is None:
        if hy_src_check(fname):
            code = _hy_code_from_file(fname, loader_type=HyLoader)
        else:
            # Try normal source
            with open(fname, "rb") as f:
                # This code differs from `runpy`'s only in that we
                # force decoding into UTF-8.
                source = f.read().decode("utf-8")
            code = compile(
                source, fname, "exec",
                **(dict(module=module) if hy.compat.PY3_15 else {}))

    return code if hy.compat.PY3_12_6 else (code, fname)


importlib.machinery.SOURCE_SUFFIXES.insert(0, ".hy")
_py_source_to_code = importlib.machinery.SourceFileLoader.source_to_code


def _could_be_hy_src(filename):
    return (
        os.path.splitext(filename)[1]
        not in set(importlib.machinery.SOURCE_SUFFIXES) - {".hy"}
    )


def _hy_source_to_code(self, data, path, fullname=None, _optimize=-1):
    if _could_be_hy_src(path):
        if os.environ.get("HY_MESSAGE_WHEN_COMPILING"):
            print("Compiling", path, file=sys.stderr)
        source = data.decode("utf-8")
        hy_tree = read_many(source, filename=path, skip_shebang=True, reader=HyReader())
        with loader_module_obj(self) as module:
            data = hy_compile(hy_tree, module)

    return _py_source_to_code(
        self, data, path,
        _optimize=_optimize,
        **(dict(fullname=fullname) if hy.compat.PY3_15 else {}))


importlib.machinery.SourceFileLoader.source_to_code = _hy_source_to_code


if (".hy", False, False) not in zipimport._zip_searchorder:
    zipimport._zip_searchorder += ((".hy", False, False),)
    _py_compile_source = zipimport._compile_source

    def _hy_compile_source(pathname, source, module=None):
        if not pathname.endswith(".hy"):
            return _py_compile_source(pathname, source, module)
        mname = f"<zip:{pathname}>"
        sys.modules[mname] = types.ModuleType(mname)
        return compile(
            hy_compile(
                read_many(source.decode("UTF-8"), filename=pathname, skip_shebang=True, reader=HyReader()),
                sys.modules[mname],
            ),
            pathname,
            "exec",
            dont_inherit=True,
        )

    zipimport._compile_source = _hy_compile_source


#  This is actually needed; otherwise, pre-created finders assigned to the
#  current dir (i.e. `''`) in `sys.path` will not catch absolute imports of
#  directory-local modules!
sys.path_importer_cache.clear()

# Do this one just in case?
importlib.invalidate_caches()


class HyLoader(importlib.machinery.SourceFileLoader):
    pass


# We create a separate version of runpy, "runhy", that prefers Hy source over
# Python.
runhy = importlib.import_module("runpy")

runhy._get_code_from_file = partial(_get_code_from_file, hy_src_check=_could_be_hy_src)

del sys.modules["runpy"]

runpy = importlib.import_module("runpy")

_runpy_get_code_from_file = runpy._get_code_from_file
runpy._get_code_from_file = _get_code_from_file


def _inject_builtins():
    """Inject the Hy core macros into Python's builtins if necessary"""
    if hasattr(builtins, "__hy_injected__"):
        return
    hy.macros.load_macros(builtins)
    # Set the marker so we don't inject again.
    builtins.__hy_injected__ = True



"""This file contains Hy's core macros that are written in Python and
return compiler Result objects (or Python AST objects) rather than Hy
model trees. These macros serve the role of special forms in other
Lisps: all ordinary macros should eventually compile down to one of
these, or to one of the model builders in hy.compiler."""

# ------------------------------------------------
# * Imports
# ------------------------------------------------

import ast
import textwrap
from contextlib import nullcontext
from itertools import dropwhile, zip_longest

from funcparserlib.parser import finished, forward_decl, many, maybe, oneplus, some

from hy import last_version
from hy.compat import PY3_11, PY3_12, PY3_15
from hy.compiler import Result, asty, mkexpr
from hy.errors import HyEvalError, HyInternalError, HyTypeError
from hy.macros import pattern_macro, require, require_reader, local_macro_name
from hy.model_patterns import (
    FORM,
    KEYWORD,
    LITERAL,
    STR,
    SYM,
    Tag,
    braces,
    brackets,
    dolike,
    in_tuple,
    keepsym,
    notpexpr,
    parse_if,
    pexpr,
    sym,
    tag,
    times,
    unpack,
)
from hy.models import (
    Bytes,
    Complex,
    Dict,
    Expression,
    FComponent,
    Float,
    FString,
    Integer,
    Keyword,
    List,
    Sequence,
    String,
    Symbol,
    Tuple,
    as_model,
    is_unpack,
)
from hy.reader import mangle, HyReader
from hy.scoping import OuterVar, ScopeFn, ScopeGen, ScopeLet, is_function_scope, is_inside_function_scope, nearest_python_scope

# ------------------------------------------------
# * Helpers
# ------------------------------------------------

Inf = float("inf")


def pvalue(root, wanted):
    return pexpr(sym(root) + wanted) >> (lambda x: x[0])


def maybe_annotated(target):
    return (
       pexpr(sym("annotate"), target, FORM).named('`annotate` form') |
       (target >> (lambda x: (x, None))))


def dotted(name):
    return Expression(map(Symbol, [".", *name.split(".")]))


type_params = sym(":tp") + brackets(many(
    maybe_annotated(SYM) | unpack("either", Symbol)))

def digest_type_params(compiler, tp):
    "Return a `type_params` attribute for `FunctionDef` etc."

    if not tp:
        return {}
    if not PY3_12:
       compiler._syntax_error(tp, "`:tp` requires Python 3.12 or later")

    return dict(type_params = [
        asty.TypeVarTuple(x[1], name = mangle(x[1]))
            if is_unpack("iterable", x) else
        asty.ParamSpec(x[1], name = mangle(x[1]))
            if is_unpack("mapping", x) else
        asty.TypeVar(x[0],
               name = mangle(x[0]),
               bound = x[1] and compiler.compile(x[1]).force_expr)
        for x in tp[0]])


# ------------------------------------------------
# * Fundamentals
# ------------------------------------------------


@pattern_macro("do", [many(FORM)])
def compile_do(compiler, expr, root, body):
    return compiler._compile_branch(body)


@pattern_macro(["eval-and-compile", "eval-when-compile", "do-mac"], [many(FORM)])
def compile_eval_foo_compile(compiler, expr, root, body):
    new_expr = Expression([Symbol("do").replace(expr[0])]).replace(expr)

    try:
        value = compiler.eval(new_expr + body)
    except HyInternalError:
        # Unexpected "meta" compilation errors need to be treated
        # like normal (unexpected) compilation errors at this level
        # (or the compilation level preceding this one).
        raise
    except Exception as e:
        # These could be expected Hy language errors (e.g. syntax errors)
        # or regular Python runtime errors that do not signify errors in
        # the compilation *process* (although compilation did technically
        # fail).
        # We wrap these exceptions and pass them through.
        raise HyEvalError(str(e), compiler.filename, body, compiler.source)

    return (
        compiler.compile(as_model(value).replace(expr))
        if root == "do-mac"
        else compiler._compile_branch(body)
        if root == "eval-and-compile"
        else Result()
    )


@pattern_macro(["py", "pys"], [STR])
def compile_inline_python(compiler, expr, root, code):
    exec_mode = root == "pys"

    try:
        o = asty.parse(
            expr,
            textwrap.dedent(code) if exec_mode else "(" + code + "\n)",
            compiler.filename,
            "exec" if exec_mode else "eval",
        ).body
    except (SyntaxError, ValueError) as e:
        compiler._syntax_error(
            expr, "Python parse error in '{}': {}".format(root, e)
        )

    return Result(stmts=o) if exec_mode else o


@pattern_macro("pragma", [many(KEYWORD + FORM)])
def compile_pragma(compiler, expr, root, kwargs):
    for kw, value in kwargs:

        if kw == Keyword("hy"):
            min_version = compiler.eval(value)
            if not isinstance(min_version, str):
                compiler._syntax_error(value, "The version given to the pragma `:hy` must be a string")
            parts = min_version.split('.')
            if not all(p.isdigit() for p in parts):
                compiler._syntax_error(value, "The string given to the pragma `:hy` must be a dot-separated sequence of integers")
            for have, need in zip_longest(
                    map(int, last_version.split('.')),
                    map(int, parts)):
                if need is None:
                    break
                if have is None or have < need:
                    compiler._syntax_error(kw, f"Hy version {min_version} or later required")
                if have > need:
                    break

        elif kw == Keyword("warn-on-core-shadow"):
            compiler.local_state_stack[-1]['warn_on_core_shadow'] = (
                bool(compiler.eval(value)))

        elif kw == Keyword("bracketed-templates"):
            reader = HyReader.current_reader(create=False)
            if reader:
                reader.bracketed_templates = bool(compiler.eval(value))

        else:
            compiler._syntax_error(kw, f"Unknown pragma `{kw}`. Perhaps it's implemented by a newer version of Hy.")

    return Result()


# ------------------------------------------------
# * Quoting
# ------------------------------------------------


@pattern_macro(["quote", "quasiquote"], [FORM])
def compile_quote(compiler, expr, root, arg):
    return compiler.compile(render_quoted_form(compiler, arg,
        level = Inf if root == "quote" else 0)[0])
          # Only quasiquotes can unquote

def render_quoted_form(compiler, form, level):
    """
    Render a quoted form as a new hy Expression.

    `level` is the level of quasiquoting of the current form. We can
    unquote if level is 0.

    Returns a two-tuple (`expression`, `splice`).

    The `splice` return value is used to mark `unquote-splice`d forms.
    We need to distinguish them as want to concatenate them instead of
    just nesting them.
    """

    op = None
    if isinstance(form, Expression) and form and isinstance(form[0], Symbol):
        op = mangle(form[0]).replace('_', '-')
        if op in ("unquote", "unquote-splice", "quasiquote"):
            if level == 0 and op != "quasiquote":
                if len(form) != 2:
                    raise HyTypeError(
                        "`%s' needs 1 argument, got %s" % op,
                        len(form) - 1,
                        compiler.filename,
                        form,
                        compiler.source,
                    )
                return form[1], op == "unquote-splice"
            level += 1 if op == "quasiquote" else -1

    name = form.__class__.__name__
    body = [form]

    if isinstance(form, Sequence):
        contents = []
        for x in form:
            f_contents, splice = render_quoted_form(compiler, x, level)
            if splice:
                if is_unpack("iterable", f_contents):
                    compiler._syntax_error(f_contents, "`unpack-iterable` is not allowed here")
                f_contents = Expression(
                    [
                        Symbol("unpack-iterable"),
                        Expression([Symbol("or"), f_contents, List()]),
                    ]
                )
            contents.append(f_contents)
        body = [List(contents)]

        if isinstance(form, FString):
            if form.brackets is not None:
                body.extend([Keyword("brackets"), String(form.brackets)])
            if form.is_tstring:
                body.extend([Keyword("is_tstring"), Symbol("True")])
        elif isinstance(form, FComponent):
            if form.conversion is not None:
                body.extend([Keyword("conversion"), String(form.conversion)])
            if form.expression is not None:
                body.extend([Keyword("expression"), String(form.expression)])
            if form.is_tstring:
                body.extend([Keyword("is_tstring"), Symbol("True")])

    elif isinstance(form, Symbol):
        body = [String(form), Keyword("from_parser"), Symbol("True")]

    elif isinstance(form, Keyword):
        body = [String(form.name), Keyword("from_parser"), Symbol("True")]

    elif isinstance(form, String):
        if form.brackets is not None:
            body.extend([Keyword("brackets"), String(form.brackets)])

    return (Expression([dotted("hy.models." + name), *body]).replace(form), False)


# ------------------------------------------------
# * Python operators
# ------------------------------------------------


@pattern_macro(["not", "bnot"], [FORM], shadow=True)
def compile_unary_operator(compiler, expr, root, arg):
    ops = {"not": ast.Not, "bnot": ast.Invert}
    operand = compiler.compile(arg)
    return operand + asty.UnaryOp(expr, op=ops[root](), operand=operand.force_expr)


@pattern_macro(["and", "or"], [many(FORM)], shadow=True)
def compile_logical_or_and_and_operator(compiler, expr, operator, args):
    ops = {"and": (ast.And, True), "or": (ast.Or, None)}
    opnode, default = ops[operator]
    if len(args) == 0:
        return asty.Constant(expr[0], value=default)

    ret = None
    var = None          # A temporary variable for assigning results to
    assignment = None   # The current assignment to `var`
    stmts = None        # The current statement list
    can_append = False  # Whether the current expression is the compiled boolop

    def put(node, value):
        # Save the result of the operation so far to `var`.
        nonlocal var, assignment, can_append
        if var is None:
            var = compiler.get_anon_var()
        name = asty.Name(node, id=var, ctx=ast.Store())
        ret.temp_variables.append(name)
        can_append = False
        return (assignment := asty.Assign(node, targets=[name], value=value))

    def get(node):
        # Get the value of `var`, creating it if necessary.
        if var is None:
            stmts.append(put(node, ret.force_expr))
        name = asty.Name(node, id=var, ctx=ast.Load())
        ret.temp_variables.append(name)
        return name

    for value in map(compiler.compile, args):
        if ret is None:
            # This is the first iteration. Don't actually introduce a
            # `BoolOp` yet; the unary case doesn't need it.
            ret = value
            stmts = ret.stmts
            can_append = False
        elif value.stmts:
            # Save the result of the statements to the temporary
            # variable. Use an `if` statement to implement
            # short-circuiting from this point.
            node = value.stmts[0]
            cond = get(node)
            if operator == "or":
                # Negate the conditional.
                cond = asty.UnaryOp(node, op=ast.Not(), operand=cond)
            branch = asty.If(node, test=cond, body=value.stmts, orelse=[])
            stmts.append(branch)
            stmts = branch.body
            stmts.append(put(node, value.force_expr))
        else:
            # Add this value to the current `BoolOp`, or create a new
            # one if we don't have one.
            value = value.force_expr
            def enbool(expr):
                nonlocal can_append
                if can_append:
                    expr.values.append(value)
                    return expr
                can_append = True
                return asty.BoolOp(expr, op=opnode(), values=[expr, value])
            if assignment:
                assignment.value = enbool(assignment.value)
            else:
                ret.expr = enbool(ret.force_expr)

    if var:
        ret.expr = get(expr)
    return ret


c_ops = {
    "=": ast.Eq,
    "!=": ast.NotEq,
    "<": ast.Lt,
    "<=": ast.LtE,
    ">": ast.Gt,
    ">=": ast.GtE,
    "is": ast.Is,
    "is-not": ast.IsNot,
    "in": ast.In,
    "not-in": ast.NotIn,
}
c_ops = {mangle(k): v for k, v in c_ops.items()}


def get_c_op(compiler, sym):
    k = mangle(sym)
    if k not in c_ops:
        compiler._syntax_error(sym, "Illegal comparison operator: " + str(sym))
    return c_ops[k]()


@pattern_macro(["=", "is", "<", "<=", ">", ">="], [oneplus(FORM)], shadow=True)
@pattern_macro(["!=", "is-not", "in", "not-in"], [times(2, Inf, FORM)], shadow=True)
def compile_compare_op_expression(compiler, expr, root, args):
    if len(args) == 1:
        return compiler.compile(args[0]) + asty.Constant(expr, value=True)

    ops = [get_c_op(compiler, root) for _ in args[1:]]
    exprs, ret, _ = compiler._compile_collect(args)
    return ret + asty.Compare(expr, left=exprs[0], ops=ops, comparators=exprs[1:])


@pattern_macro("chainc", [FORM, many(SYM + FORM)])
def compile_chained_comparison(compiler, expr, root, arg1, args):
    ret = compiler.compile(arg1)
    arg1 = ret.force_expr

    ops = [get_c_op(compiler, op) for op, _ in args]
    args, ret2, _ = compiler._compile_collect([x for _, x in args])

    return ret + ret2 + asty.Compare(expr, left=arg1, ops=ops, comparators=args)


# The second element of each tuple below is an aggregation operator
# that's used for augmented assignment with three or more arguments.
m_ops = {
    "+": (ast.Add, "+"),
    "/": (ast.Div, "*"),
    "//": (ast.FloorDiv, "*"),
    "*": (ast.Mult, "*"),
    "-": (ast.Sub, "+"),
    "%": (ast.Mod, None),
    "**": (ast.Pow, "**"),
    "<<": (ast.LShift, "+"),
    ">>": (ast.RShift, "+"),
    "|": (ast.BitOr, "|"),
    "^": (ast.BitXor, None),
    "&": (ast.BitAnd, "&"),
    "@": (ast.MatMult, "@"),
}


@pattern_macro(["+", "*", "|"], [many(FORM)], shadow=True)
@pattern_macro(["-", "/", "&", "@"], [oneplus(FORM)], shadow=True)
@pattern_macro(["**", "//", "<<", ">>"], [times(2, Inf, FORM)], shadow=True)
@pattern_macro(["%", "^"], [times(2, 2, FORM)], shadow=True)
def compile_maths_expression(compiler, expr, root, args):
    if len(args) == 0:
        # Return the identity element for this operator.
        return asty.Constant(expr, value=({"+": 0, "|": 0, "*": 1}[root]))

    if len(args) == 1:
        if root == "/":
            # Compute the reciprocal of the argument.
            args = [Integer(1).replace(expr), args[0]]
        elif root in ("+", "-"):
            # Apply unary plus or unary minus to the argument.
            op = {"+": ast.UAdd, "-": ast.USub}[root]()
            ret = compiler.compile(args[0])
            return ret + asty.UnaryOp(expr, op=op, operand=ret.force_expr)
        else:
            # Return the argument unchanged.
            return compiler.compile(args[0])

    op = m_ops[root][0]
    right_associative = root == "**"
    ret = compiler.compile(args[-1 if right_associative else 0])
    for child in args[-2 if right_associative else 1 :: -1 if right_associative else 1]:
        left_expr = ret.force_expr
        ret += compiler.compile(child)
        right_expr = ret.force_expr
        if right_associative:
            left_expr, right_expr = right_expr, left_expr
        ret += asty.BinOp(expr, left=left_expr, op=op(), right=right_expr)

    return ret


a_ops = {x + "=": v for x, v in m_ops.items()}


@pattern_macro(
    [x for x, (_, v) in a_ops.items() if v is not None], [FORM, oneplus(FORM)]
)
@pattern_macro(
    [x for x, (_, v) in a_ops.items() if v is None], [FORM, times(1, 1, FORM)]
)
def compile_augassign_expression(compiler, expr, root, target, values):
    if len(values) > 1:
        return compiler.compile(
            mkexpr(root, [target], mkexpr(a_ops[root][1], rest=values)).replace(expr)
        )

    op = a_ops[root][0]
    target = compiler._storeize(target, compiler.compile(target))
    ret = compiler.compile(values[0])
    return ret + asty.AugAssign(expr, target=target, value=ret.force_expr, op=op())


# ------------------------------------------------
# * Assignment, mutation, and annotation
# ------------------------------------------------


@pattern_macro("setv", [many(
    (tag("chained", sym(":chain") + brackets(oneplus(FORM))) |
        tag("maybe_annotated", maybe_annotated(FORM))) +
    FORM)])
@pattern_macro("setx", [times(1, 1, tag("setx", SYM) + FORM)])
def compile_def_expression(compiler, expr, root, decls):
    if not decls:
        return asty.Constant(expr, value=None)

    result = Result()

    for (ttag, target), value in decls:
        ann = None
        if ttag == "chained":
            [target] = target
        if ttag == "maybe_annotated":
            target, ann = target
        result += compile_assign(
            compiler, ann, target, value,
            is_assignment_expr = ttag == "setx",
            chained = ttag == "chained"
        )
    return result


@pattern_macro("let", [brackets(many(maybe_annotated(FORM) + FORM)), many(FORM)])
def compile_let(compiler, expr, root, bindings, body):
    res = Result()
    bindings = bindings[0]
    scope = compiler.scope.create(ScopeLet)

    for (target, ann), value in bindings:
        res += compile_assign(compiler, ann, target, value, let_scope=scope)

    with scope:
        return res + compiler.compile(mkexpr("do", *body).replace(expr))


@pattern_macro(["annotate"], [FORM, FORM])
def compile_basic_annotation(compiler, expr, root, target, ann):
    return compile_assign(compiler, ann, target, None)


def compile_assign(
    compiler, ann, target, value, *, is_assignment_expr=False, chained=False, let_scope=None
):
    # Ensure that assignment expressions have a result and no annotation.
    assert not is_assignment_expr or (value is not None and ann is None)

    annotate_only = value is None
    if annotate_only:
        result = Result()
    else:
        with let_scope or nullcontext():
            result = compiler.compile(value)
        if let_scope:
            target = let_scope.add(target)

    if result.temp_variables and isinstance(target, Symbol):
        result.rename(compiler, compiler._nonconst(target))
        if not is_assignment_expr:
            # Throw away .expr to ensure that (setv ...) returns None.
            result.expr = None
    else:
        st_targets = [
           compiler._storeize(t, compiler.compile(t))
           for t in (target if chained else [target])]

        if ann is not None:
            ann_result = compiler.compile(ann)
            result = ann_result + result

        target_kwarg = dict(target = st_targets[0])
        if is_assignment_expr:
            node = asty.NamedExpr
        elif ann is not None:
            node = lambda x, **kw: asty.AnnAssign(
                x,
                annotation=ann_result.force_expr,
                simple=int(isinstance(target, Symbol)),
                **kw,
            )
        else:
            node = asty.Assign
            target_kwarg = dict(targets = st_targets)

        result += node(
            target if hasattr(target, "start_line") else result,
            value=result.force_expr if not annotate_only else None,
            **target_kwarg
        )

    return result


@pattern_macro(((3, 12), "deftype"), [maybe(type_params), SYM, FORM])
def compile_deftype(compiler, expr, root, tp, name, value):
    return asty.TypeAlias(expr,
       name = asty.Name(name, id = mangle(name), ctx = ast.Store()),
       value = compiler.compile(value).force_expr,
        **digest_type_params(compiler, tp))


@pattern_macro(["global", "nonlocal"], [many(SYM)])
def compile_global_or_nonlocal(compiler, expr, root, syms):
    if not syms:
        return asty.Pass(expr)

    names = [mangle(s) for s in syms]
    ret = (asty.Global(expr, names = names)
        if root == "global" else
        OuterVar(expr, compiler.scope, names))

    try:
        compiler.scope.define_nonlocal(ret, root)
    except SyntaxError as e:
        compiler._syntax_error(expr, e.msg)

    return ret


@pattern_macro("del", [many(FORM)])
def compile_del_expression(compiler, expr, name, args):
    if not args:
        return asty.Pass(expr)

    del_targets = []
    ret = Result()
    for target in args:
        compiled_target = compiler.compile(target)
        ret += compiled_target
        del_targets.append(compiler._storeize(target, compiled_target, ast.Del))

    return ret + asty.Delete(expr, targets=del_targets)


# ------------------------------------------------
# * Subsetting
# ------------------------------------------------


@pattern_macro("get", [FORM, oneplus(FORM)], shadow=True)
def compile_index_expression(compiler, expr, name, obj, indices):
    indices, ret, _ = compiler._compile_collect(indices)
    ret += compiler.compile(obj)

    for ix in indices:
        ret += asty.Subscript(
            expr, value=ret.force_expr, slice=ast.Index(value=ix), ctx=ast.Load()
        )

    return ret


notsym = lambda *dissallowed: some(
    lambda x: isinstance(x, Symbol) and str(x) not in dissallowed
)


@pattern_macro(
    ".",
    [
        FORM,
        many(
            SYM
            | brackets(FORM)
            | pexpr(notsym("unpack-iterable", "unpack-mapping"), many(FORM))
        ),
    ],
)
def compile_attribute_access(compiler, expr, name, invocant, keys):
    ret = compiler.compile(invocant)

    for attr in keys:
        if isinstance(attr, Symbol):
            ret += asty.Attribute(
                attr, value=ret.force_expr, attr=mangle(attr), ctx=ast.Load()
            )
        elif isinstance(attr, Expression):
            root, args = attr
            func = asty.Attribute(
                root, value=ret.force_expr, attr=mangle(root), ctx=ast.Load()
            )

            args, funcret, keywords = compiler._compile_collect(args, with_kwargs=True)
            ret += (
                funcret
                + func
                + asty.Call(expr, func=func, args=args, keywords=keywords)
            )
        else:  # attr is a hy List
            compiled_attr = compiler.compile(attr[0])
            ret = (
                compiled_attr
                + ret
                + asty.Subscript(
                    attr,
                    value=ret.force_expr,
                    slice=ast.Index(value=compiled_attr.force_expr),
                    ctx=ast.Load(),
                )
            )

    return ret


@pattern_macro("cut", [FORM, maybe(FORM), maybe(FORM), maybe(FORM)])
def compile_cut_expression(compiler, expr, name, obj, lower, upper, step):
    ret = [Result()]

    def c(e):
        ret[0] += compiler.compile(e)
        return ret[0].force_expr

    if upper is None:
        # cut with single index is an upper bound,
        # this is consistent with slice and islice
        upper = lower
        lower = Symbol("None")

    s = asty.Subscript(
        expr,
        value=c(obj),
        slice=asty.Slice(expr, lower=c(lower), upper=c(upper), step=c(step)),
        ctx=ast.Load(),
    )
    return ret[0] + s


@pattern_macro("unpack-iterable", [FORM])
def compile_unpack_iterable(compiler, expr, root, arg):
    ret = compiler.compile(arg)
    ret += asty.Starred(expr, value=ret.force_expr, ctx=ast.Load())
    return ret


# ------------------------------------------------
# * `if`
# ------------------------------------------------


@pattern_macro("if", [FORM, FORM, FORM])
def compile_if(compiler, expr, _, cond, body, orel_expr):
    cond = compiler.compile(cond)
    body = compiler.compile(body)

    nested = root = False
    orel = Result()
    if (
        isinstance(orel_expr, Expression)
        and isinstance(orel_expr[0], Symbol)
        and orel_expr[0] == Symbol("if*")
    ):
        # Nested ifs: don't waste temporaries
        root = compiler.temp_if is None
        nested = True
        compiler.temp_if = compiler.temp_if or compiler.get_anon_var()
    orel = compiler.compile(orel_expr)

    if not cond.stmts and isinstance(cond.force_expr, ast.Name):
        name = cond.force_expr.id
        branch = None
        if name == "True":
            branch = body
        elif name in ("False", "None"):
            branch = orel
        if branch is not None:
            if compiler.temp_if and branch.stmts:
                name = asty.Name(expr, id=mangle(compiler.temp_if), ctx=ast.Store())

                branch += asty.Assign(expr, targets=[name], value=body.force_expr)

            return branch

    # We want to hoist the statements from the condition
    ret = cond

    if body.stmts or orel.stmts:
        # We have statements in our bodies
        # Get a temporary variable for the result storage
        var = compiler.temp_if or compiler.get_anon_var()
        name = asty.Name(expr, id=mangle(var), ctx=ast.Store())

        # Store the result of the body
        body += asty.Assign(expr, targets=[name], value=body.force_expr)

        # and of the else clause
        if not nested or not orel.stmts or (not root and var != compiler.temp_if):
            orel += asty.Assign(expr, targets=[name], value=orel.force_expr)

        # Then build the if
        ret += asty.If(expr, test=ret.force_expr, body=body.stmts, orelse=orel.stmts)

        # And make our expression context our temp variable
        expr_name = asty.Name(expr, id=mangle(var), ctx=ast.Load())

        ret += Result(expr=expr_name, temp_variables=[expr_name, name])
    else:
        # Just make that an if expression
        ret += asty.IfExp(
            expr, test=ret.force_expr, body=body.force_expr, orelse=orel.force_expr
        )

    if root:
        compiler.temp_if = None

    return ret


# ------------------------------------------------
# * The `for` family
# ------------------------------------------------

loopers = many(
    tag("setv", sym(":setv") + FORM + FORM)
    | tag("if", sym(":if") + FORM)
    | tag("do", sym(":do") + FORM)
    | tag("afor", sym(":async") + FORM + FORM)
    | tag("for", FORM + FORM)
)


@pattern_macro(["for"], [
    brackets(loopers, name = 'square-bracketed loop clauses'),
    many(notpexpr("else")) + maybe(dolike("else"))])
@pattern_macro(["lfor", "sfor", "gfor"], [loopers, FORM])
@pattern_macro(["dfor"], [loopers, finished | unpack("mapping")])
# Here `finished` is a hack replacement for FORM + FORM:
# https://github.com/vlasovskikh/funcparserlib/issues/75
def compile_comprehension(compiler, expr, root, parts, final):
    node_class = {
        "for": asty.For,
        "lfor": asty.ListComp,
        "dfor": asty.DictComp,
        "sfor": asty.SetComp,
        "gfor": asty.GeneratorExp,
    }[root]
    is_for = root == "for"

    ctx = nullcontext() if is_for else compiler.scope.create(ScopeGen)
    mac_con = nullcontext() if is_for else compiler.local_state()
    with mac_con, ctx as scope:

        # Compile the parts.
        dict_unpack = False
        if is_for:
            parts = parts[0]
        if node_class is asty.DictComp:
            dict_unpack = bool(final)
            if not (dict_unpack or (parts and parts[-1].tag == "for")):
                compiler._syntax_error(
                    parts[-1] if parts else parts,
                    "`dfor` must end with key and value forms, or `#** FORM`",
                )
            final = final or parts.pop().value
        if not parts:
            return Result(
                expr=asty.parse(
                    expr,
                    {
                        asty.For: "None",
                        asty.ListComp: "[]",
                        asty.DictComp: "{}",
                        asty.SetComp: "{1}.__class__()",
                        asty.GeneratorExp: "(_ for _ in [])",
                    }[node_class],
                )
                .body[0]
                .value
            )
        new_parts = []
        for p in parts:
            if p.tag in ("if", "do"):
                tag_value = compiler.compile(p.value)
            else:
                tag_value = [
                    compiler._storeize(p.value[0], compiler.compile(p.value[0])),
                    compiler.compile(p.value[1]),
                ]
                if not is_for:
                    scope.iterator(tag_value[0])
            new_parts.append(Tag(p.tag, tag_value))
        parts = new_parts

        orel = []
        if is_for:
            # Get the `else`.
            body, else_expr = final
            if else_expr is not None:
                orel.append(compiler._compile_branch(else_expr))
                orel[0] += orel[0].expr_as_stmt()
        else:
            # Get the final value (and for dictionary
            # comprehensions, the final key).
            key = elt = None
            if node_class is asty.DictComp:
                if dict_unpack:
                    key = compiler.compile(final[1])
                else:
                    key, elt = map(compiler.compile, final)
            else:
                elt = compiler.compile(final)

        ends_with_unpack = not is_for and (dict_unpack or (elt and isinstance(elt.expr, ast.Starred)))

        # Produce a result.
        if (
            is_for
            or (elt is not None and elt.stmts)
            or (key is not None and key.stmts)
            or (not PY3_15 and ends_with_unpack)
            or any(
                p.tag == "do"
                or (
                    p.value[1].stmts
                    if p.tag in ("for", "afor", "setv")
                    else p.value.stmts
                )
                for p in parts
            )
        ):
            # The desired comprehension can't be expressed as a
            # real Python comprehension. We'll write it as a nested
            # loop in a function instead.
            any_async = False
            def f(parts):
                # This function is called recursively to construct
                # the nested loop.
                nonlocal elt, ends_with_unpack, any_async
                if not parts:
                    if is_for:
                        if body:
                            bd = compiler._compile_branch(body)
                            return bd + bd.expr_as_stmt()
                        return Result(stmts=[asty.Pass(expr)])
                    if ends_with_unpack:
                        ends_with_unpack = False
                        to_loop = Result(expr =
                            # Call `key.items()` so we yield (key,
                            # value) pairs instead of just keys.
                            asty.Call(key,
                                args = [],
                                keywords = [],
                                func = asty.Attribute(key,
                                    value = key.force_expr,
                                    attr = 'items',
                                    ctx = ast.Load()))
                            if dict_unpack
                            else elt.expr.value)
                        temp_var = compiler.get_anon_var()
                        elt = compiler.compile(Symbol(temp_var).replace(final))
                        return f([["for", (
                            asty.Name(final, id=mangle(temp_var), ctx=ast.Store()),
                            to_loop)]])
                    if node_class is asty.DictComp and not dict_unpack:
                        ret = key + elt
                        val = asty.Tuple(
                            key, ctx=ast.Load(), elts=[key.force_expr, elt.force_expr]
                        )
                    else:
                        ret = elt
                        val = elt.force_expr
                    return ret + asty.Expr(elt, value=asty.Yield(elt, value=val))
                (tagname, v), parts = parts[0], parts[1:]
                if tagname in ("for", "afor"):
                    orelse = orel and orel.pop().stmts
                    node = asty.AsyncFor if tagname == "afor" else asty.For
                    any_async = any_async or tagname == "afor"
                    return v[1] + node(
                        v[1],
                        target=v[0],
                        iter=v[1].force_expr,
                        body=f(parts).stmts,
                        orelse=orelse,
                    )
                elif tagname == "setv":
                    return (
                        v[1]
                        + asty.Assign(v[1], targets=[v[0]], value=v[1].force_expr)
                        + f(parts)
                    )
                elif tagname == "if":
                    return v + asty.If(
                        v, test=v.force_expr, body=f(parts).stmts, orelse=[]
                    )
                elif tagname == "do":
                    return v + v.expr_as_stmt() + f(parts)
                else:
                    raise ValueError("can't happen")

            if is_for:
                return f(parts)
            fname = compiler.get_anon_var()
            # Define the generator function.
            stmts = []
            ret = Result()
            assignment_names = scope.finalize()
            if scope.exposing_assignments and assignment_names:
                # expose inner assignments to outer scope
                unlocal_type = (
                    asty.Nonlocal
                    if is_inside_function_scope(scope.parent)
                    else asty.Global
                )
                stmts.append(unlocal_type(expr, names=assignment_names))

                # create a fake assignment statement so python places these
                # names in the immediately outer scope
                if_body = []
                if scope.nonlocal_vars:
                    if_body.append(
                        asty.Nonlocal(expr, names=list(sorted(scope.nonlocal_vars)))
                    )
                assignments = asty.Tuple(
                    expr,
                    elts=[
                        asty.Name(expr, id=var, ctx=ast.Store())
                        for var in assignment_names
                    ],
                    ctx=ast.Store(),
                )
                if_body.append(
                    asty.Assign(
                        expr,
                        targets=[assignments],
                        value=asty.Constant(expr, value=None),
                    )
                )
                ret += asty.If(
                    expr, test=asty.Constant(expr, value=False), body=if_body, orelse=[]
                )

            body = f(parts).stmts
              # `f` needs to be called before the next line so
              # `any_async` is set early enough.
            ret += (asty.AsyncFunctionDef if any_async else asty.FunctionDef)(
                expr,
                name=fname,
                args=ast.arguments(
                    args=[],
                    vararg=None,
                    kwarg=None,
                    posonlyargs=[],
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                ),
                body=stmts + body,
                decorator_list=[],
                **({"type_params": []} if PY3_12 else {}),
            )
            # Immediately call the new function. Unless the user asked
            # for a generator, wrap the call in another comprehension
            # to get the right type. We don't want to just use e.g.
            # `list(...)` because the name `list` might be rebound, and it
            # doesn't work with async generators.
            brackets = "[]" if node_class is asty.ListComp else "{}"
            if node_class is asty.DictComp:
                v1, v2 = compiler.get_anon_var(), compiler.get_anon_var()
                v1, v2 = f"{v1}: {v2}", f"{v1}, {v2}"
            else:
                v1 = v2 = compiler.get_anon_var()
            return ret + Result(expr =
                asty.parse(expr,
                    f"{fname}()"
                    if node_class is asty.GeneratorExp else
                    "{}{} {} for {} in {}(){}".format(
                        brackets[0],
                        v1,
                        "async" if any_async else "",
                        v2,
                        fname,
                        brackets[1]))
                .body[0].value)

        # We can produce a real comprehension.
        generators = []
        for tagname, v in parts:
            if tagname in ("for", "afor"):
                generators.append(
                    ast.comprehension(
                        target=v[0],
                        iter=v[1].expr,
                        ifs=[],
                        is_async=int(tagname == "afor"),
                    )
                )
            elif tagname == "setv":
                generators.append(
                    ast.comprehension(
                        target=v[0],
                        iter=asty.Tuple(v[1], elts=[v[1].expr], ctx=ast.Load()),
                        ifs=[],
                        is_async=0,
                    )
                )
            elif tagname == "if":
                generators[-1].ifs.append(v.expr)
            else:
                raise ValueError("can't happen")
        if node_class is asty.DictComp:
            return asty.DictComp(
                expr, key=key.expr, value=(elt and elt.expr), generators=generators
            )
        return node_class(expr, elt=elt.expr, generators=generators)


# ------------------------------------------------
# * More looping
# ------------------------------------------------


@pattern_macro(["while"], [FORM, many(notpexpr("else")), maybe(dolike("else"))])
def compile_while_expression(compiler, expr, root, cond, body, else_expr):
    cond_compiled = compiler.compile(cond)

    body = compiler._compile_branch(body)
    body += body.expr_as_stmt()
    body_stmts = body.stmts or [asty.Pass(expr)]

    if cond_compiled.stmts:
        # We need to ensure the statements for the condition are
        # executed on every iteration. Rewrite the loop to use a
        # single anonymous variable as the condition, i.e.:
        #  anon_var = True
        #  while anon_var:
        #    condition stmts...
        #    anon_var = condition expr
        #    if anon_var:
        #      while loop body
        cond_var = asty.Name(cond, id=compiler.get_anon_var(), ctx=ast.Load())

        def make_not(operand):
            return asty.UnaryOp(cond, op=ast.Not(), operand=operand)

        body_stmts = cond_compiled.stmts + [
            asty.Assign(
                cond,
                targets=[compiler._storeize(cond, cond_var)],
                # Cast the condition to a bool in case it's mutable and
                # changes its truth value, but use (not (not ...)) instead of
                # `bool` in case `bool` has been redefined.
                value=make_not(make_not(cond_compiled.force_expr)),
            ),
            asty.If(cond, test=cond_var, body=body_stmts, orelse=[]),
        ]

        cond_compiled = (
            Result()
            + asty.Assign(
                cond,
                targets=[compiler._storeize(cond, cond_var)],
                value=asty.Constant(cond, value=True),
            )
            + cond_var
        )

    orel = Result()
    if else_expr is not None:
        orel = compiler._compile_branch(else_expr)
        orel += orel.expr_as_stmt()

    ret = cond_compiled + asty.While(
        expr, test=cond_compiled.force_expr, body=body_stmts, orelse=orel.stmts
    )

    return ret


@pattern_macro(["break", "continue"], [])
def compile_break_or_continue_expression(compiler, expr, root):
    return (asty.Break if root == "break" else asty.Continue)(expr)


# ------------------------------------------------
# * `with`
# ------------------------------------------------


@pattern_macro("with", [
    brackets(oneplus(maybe(keepsym(":async")) + FORM + FORM)) |
        brackets((maybe(keepsym(":async")) + FORM) >> (lambda x: [(x[0], Symbol("_"), x[1])])),
    many(FORM)])
def compile_with_expression(compiler, expr, root, args, body):

    # We'll store the result of the body in a tempvar
    temp_var = compiler.get_anon_var()
    name = asty.Name(expr, id=mangle(temp_var), ctx=ast.Store())
    # Initialize the tempvar to None in case the `with` exits
    # early with an exception.
    initial_assign = asty.Assign(
        expr, targets=[name], value=asty.Constant(expr, value=None)
    )

    [args] = args
    ret = Result(stmts=[initial_assign])
    items = []
    was_async = None
    cbody = None
    for i, (is_async, variable, ctx) in enumerate(args):
        is_async = bool(is_async)
        if was_async is None:
            was_async = is_async
        elif is_async != was_async:
            # We're compiling a `with` that mixes synchronous and
            # asynchronous context managers. Python doesn't support
            # this directly, so start a new `with` inside the body.
            cbody = compile_with_expression(compiler, expr, root, [args[i:]], body)
            break
        if not isinstance(ctx, Result):
            # In a non-recursive call, `ctx` has not yet been compiled,
            # and thus is not yet a `Result`.
            ctx = compiler.compile(ctx)
            if i == 0:
                ret += ctx
            elif ctx.stmts:
                # We need to include some statements as part of this
                # context manager, but this `with` already has at
                # least one prior context manager. So, put our
                # statements in the body and then start a new `with`.
                cbody = ctx + compile_with_expression(
                    compiler,
                    expr,
                    root,
                    [((is_async, variable, ctx), *args[i + 1:])],
                    body)
                break
        variable = (
            None
            if variable == Symbol("_")
            else compiler._storeize(variable, compiler.compile(variable))
        )
        items.append(
            ast.withitem(context_expr=ctx.force_expr, optional_vars=variable)
        )

    if not cbody:
        cbody = compiler._compile_branch(body)
        cbody += asty.Assign(expr, targets=[name], value=cbody.force_expr)

    node = asty.AsyncWith if was_async else asty.With
    ret += node(expr, body=cbody.stmts, items=items)

    # And make our expression context our temp variable
    expr_name = asty.Name(expr, id=mangle(temp_var), ctx=ast.Load())

    ret += Result(expr=expr_name)
    # We don't give the Result any temp_vars because we don't want
    # Result.rename to touch `name`. Otherwise, initial_assign will
    # clobber any preexisting value of the renamed-to variable.

    return ret


# ------------------------------------------------
# * `match`
# ------------------------------------------------
_pattern = forward_decl()
_pattern.define(
    (
        SYM
        | KEYWORD
        | LITERAL
        | brackets(many(_pattern | unpack("iterable")))
        | in_tuple(many(_pattern | unpack("iterable")))
        | pexpr(keepsym("."), many(SYM))
        | pexpr(keepsym("|"), many(_pattern))
        | braces(many(LITERAL + _pattern), maybe(pvalue("unpack-mapping", SYM)))
        | pexpr(
            pexpr(keepsym("."), oneplus(SYM))
            | notsym(".", "|", "unpack-mapping", "unpack-iterable"),
            many(parse_if(lambda x: not isinstance(x, Keyword), _pattern)),
            many(KEYWORD + _pattern),
        )
    )
    + maybe(sym(":as") + SYM)
)
match_clause = _pattern + maybe(sym(":if") + FORM)


@pattern_macro(((3, 10), "match"), [FORM, many(match_clause + FORM)])
def compile_match_expression(compiler, expr, root, subject, clauses):
    subject = compiler.compile(subject)
    return_var = asty.Name(expr, id=mangle(compiler.get_anon_var()), ctx=ast.Store())

    lifted_if_defs = []
    match_cases = []
    for *pattern, guard, body in clauses:
        if guard and body == Keyword("as"):
            compiler._syntax_error(body, ":as clause cannot come after :if guard")

        body = compiler._compile_branch([body])
        body += asty.Assign(pattern[0], targets=[return_var], value=body.force_expr)
        body += body.expr_as_stmt()
        body = body.stmts

        pattern = compile_pattern(compiler, pattern)

        if guard:
            guard = compiler.compile(guard)
            if guard.stmts:
                fname = compiler.get_anon_var()
                guardret = Result() + asty.FunctionDef(
                    guard,
                    name=fname,
                    args=ast.arguments(
                        args=[],
                        kwarg=None,
                        posonlyargs=[],
                        kwonlyargs=[],
                        kw_defaults=[],
                        defaults=[],
                    ),
                    body=guard.stmts + [asty.Return(guard.expr, value=guard.expr)],
                    decorator_list=[],
                    **({"type_params": []} if PY3_12 else {}),
                )
                lifted_if_defs.append(guardret)
                guard = Result(expr=asty.parse(guard, f"{fname}()").body[0].value)

        match_cases.append(
            ast.match_case(
                pattern=pattern,
                guard=guard.force_expr if guard else None,
                body=body,
            )
        )

    returnable = Result(
        expr=asty.Name(expr, id=return_var.id, ctx=ast.Load()),
        temp_variables=[return_var],
    )
    ret = Result() + subject
    ret += asty.Assign(
        expr, targets=[return_var], value=asty.Constant(expr, value=None)
    )
    if not match_cases:
        return ret + returnable

    for lifted_if in lifted_if_defs:
        ret += lifted_if
    ret += asty.Match(expr, subject=subject.force_expr, cases=match_cases)
    return ret + returnable


def compile_pattern(compiler, pattern):
    value, assignment = pattern
    if assignment is not None:
        return compiler.scope.assign(
            asty.MatchAs(
                value,
                pattern=compile_pattern(compiler, (value, None)),
                name=mangle(compiler._nonconst(assignment)),
            )
        )

    if str(value) in ("None", "True", "False"):
        return asty.MatchSingleton(
            value,
            value=compiler.compile(value).force_expr.value,
        )
    elif isinstance(value, (String, Integer, Float, Complex, Bytes)):
        return asty.MatchValue(
            value,
            value=compiler.compile(value).expr,
        )
    elif value == Symbol("_"):
        return asty.MatchAs(value)
    elif isinstance(value, Symbol):
        return compiler.scope.assign(asty.MatchAs(value, name=mangle(value)))
    elif isinstance(value, Expression) and value[0] == Symbol("|"):
        return asty.MatchOr(
            value,
            patterns=[compile_pattern(compiler, v) for v in value[1]],
        )
    elif isinstance(value, Expression) and value[0] == Symbol("."):
        root, syms = value
        dotform = mkexpr(root, *syms).replace(value)
        return asty.MatchValue(
            value,
            value=compiler.compile(dotform).expr,
        )
    elif isinstance(value, (Tuple, List)):
        patterns = value[0]
        patterns = [
            compile_pattern(compiler, (v, None) if is_unpack("iterable", v) else v)
            for v in patterns
        ]
        return asty.MatchSequence(value, patterns=patterns)
    elif is_unpack("iterable", value):
        return compiler.scope.assign(asty.MatchStar(value, name=mangle(value[1])))

    elif isinstance(value, Dict):
        kvs, rest = value
        keys, values = zip(*kvs) if kvs else ([], [])
        # Call `scope.assign` for the assignment to `rest`.
        return compiler.scope.assign(
            asty.MatchMapping(
                value,
                keys=[compiler.compile(key).expr for key in keys],
                patterns=[compile_pattern(compiler, v) for v in values],
                rest=mangle(rest) if rest else None,
            )
        )
    elif isinstance(value, Expression):
        head, args, kwargs = value
        keywords, values = zip(*kwargs) if kwargs else ([], [])
        return asty.MatchClass(
            value,
            cls=compiler.compile(
              # `head` could be a symbol or a dotted form.
                (head[:1] + head[1]).replace(head)
                if type(head) is Expression
                else head).expr,
            patterns=[compile_pattern(compiler, v) for v in args],
            kwd_attrs=[kwd.name for kwd in keywords],
            kwd_patterns=[compile_pattern(compiler, value) for value in values],
        )
    elif isinstance(value, Keyword):
        return asty.MatchClass(
            value,
            cls=compiler.compile(dotted("hy.models.Keyword")).expr,
            patterns=[
                asty.MatchValue(value, value=asty.Constant(value, value=value.name))
            ],
            kwd_attrs=[],
            kwd_patterns=[],
        )
    else:
        compiler._syntax_error(value, "unsupported")


# ------------------------------------------------
# * `raise` and `try`
# ------------------------------------------------


@pattern_macro("raise", [maybe(FORM), maybe(sym(":from") + FORM)])
def compile_raise_expression(compiler, expr, root, exc, cause):
    ret = Result()

    if exc is not None:
        exc = compiler.compile(exc)
        ret += exc
        exc = exc.force_expr

    if cause is not None:
        cause = compiler.compile(cause)
        ret += cause
        cause = cause.force_expr

    return ret + asty.Raise(expr, exc=exc, cause=cause)


@pattern_macro(
    "try",
    [
        many(notpexpr("except", "except*", "else", "finally")),
        many(
            pexpr(
                keepsym("except") | keepsym("except*"),
                brackets() | brackets(FORM) | brackets(SYM, FORM),
                many(FORM),
            )
        ),
        maybe(dolike("else")),
        maybe(dolike("finally")),
    ],
)
def compile_try_expression(compiler, expr, root, body, catchers, orelse, finalbody):
    if orelse is not None and not catchers:
        # Python forbids `else` when there are no `except` clauses.
        # But we can get the same effect by appending the `else` forms
        # to the body.
        body += list(orelse)
        orelse = None
    body = compiler._compile_branch(body)
    if not (catchers or finalbody):
        # Python forbids this, so just return the body, per `do`.
        return body

    return_var = asty.Name(expr, id=mangle(compiler.get_anon_var()), ctx=ast.Store())

    handler_results = Result()
    handlers = []
    except_syms_seen = set()
    for catcher in catchers:
        # exceptions catch should be either:
        # [[list of exceptions]]
        # or
        # [variable [list of exceptions]]
        # or
        # [variable exception]
        # or
        # [exception]
        # or
        # []
        except_sym, exceptions, ebody = catcher
        if not PY3_11 and except_sym == Symbol("except*"):
            compiler._syntax_error(except_sym, "`{}` requires Python 3.11 or later")
        except_syms_seen.add(str(except_sym))
        if len(except_syms_seen) > 1:
            compiler._syntax_error(
                except_sym, "cannot have both `except` and `except*` on the same `try`"
            )

        name = None
        if len(exceptions) == 0:
            exceptions = "ALL"
        elif len(exceptions) == 1:
            [exceptions] = exceptions
        else:
            [name, exceptions] = exceptions
            name = mangle(compiler._nonconst(name))

        if exceptions == "ALL":
            # Catch all exceptions.
            types = Result()
        elif isinstance(exceptions, List):
            # [FooBar BarFoo] → Catch Foobar and BarFoo exceptions.
            elts, types, _ = compiler._compile_collect(exceptions)
            types += asty.Tuple(exceptions, elts=elts, ctx=ast.Load())
        else:
            # A single exception type.
            types = compiler.compile(exceptions)

        # Create a "fake" scope for the exception variable.
        # See: https://docs.python.org/3/reference/compound_stmts.html#the-try-statement
        with compiler.scope.create(ScopeLet) as scope:
            if name:
                name = scope.add(name, compiler.get_anon_var("exc", name))
            ebody = compiler._compile_branch(ebody)
        ebody += asty.Assign(catcher, targets=[return_var], value=ebody.force_expr)
        ebody += ebody.expr_as_stmt()

        handler_results += types + asty.ExceptHandler(
            catcher,
            type=types.expr,
            name=name,
            body=ebody.stmts or [asty.Pass(catcher)],
        )
        handlers.append(handler_results.stmts.pop())

    if orelse is None:
        orelse = []
    else:
        orelse = compiler._compile_branch(orelse)
        orelse += asty.Assign(expr, targets=[return_var], value=orelse.force_expr)
        orelse += orelse.expr_as_stmt()
        orelse = orelse.stmts

    if finalbody is None:
        finalbody = []
    else:
        finalbody = compiler._compile_branch(finalbody)
        finalbody += finalbody.expr_as_stmt()
        finalbody = finalbody.stmts

    returnable = Result(
        expr=asty.Name(expr, id=return_var.id, ctx=ast.Load()),
        temp_variables=[return_var],
    )
    body += (
        body.expr_as_stmt()
        if orelse
        else asty.Assign(expr, targets=[return_var], value=body.force_expr)
    )
    body = body.stmts or [asty.Pass(expr)]

    x = (asty.TryStar if "except*" in except_syms_seen else asty.Try)(
        expr, body=body, handlers=handlers, orelse=orelse, finalbody=finalbody
    )
    return handler_results + x + returnable


# ------------------------------------------------
# * Functions and macros
# ------------------------------------------------

NASYM = some(lambda x: isinstance(x, Symbol) and x not in (Symbol("/"), Symbol("*")))
argument = maybe_annotated(NASYM | brackets(NASYM, FORM))
varargs = lambda unpack_type, wanted: maybe_annotated(pvalue(unpack_type, wanted))
kwonly_delim = some(lambda x: x == Symbol("*"))
lambda_list = brackets(
    maybe(many(argument) + sym("/")),
    many(argument),
    maybe(kwonly_delim | varargs("unpack-iterable", NASYM)),
    many(argument),
    maybe(varargs("unpack-mapping", NASYM)),
)


@pattern_macro("fn", [
    maybe(keepsym(":async")),
    maybe(type_params),
    maybe_annotated(lambda_list),
    many(FORM)])
def compile_function_lambda(compiler, expr, root, is_async, tp, params, body):
    params, returns = params
    posonly, args, rest, kwonly, kwargs = params
    has_annotations = returns is not None or any(
        isinstance(param, tuple) and param[1] is not None
        for param in (posonly or []) + args + kwonly + [rest, kwargs]
    )
    args, ret = compile_lambda_list(compiler, params)
    with compiler.local_state(), compiler.scope.create(ScopeFn, args, is_async) as scope:
        body = compiler._compile_branch(body)

    # Compile to lambda if we can
    if not (has_annotations or tp or body.stmts or is_async):
        return ret + asty.Lambda(expr, args=args, body=body.force_expr)

    # Otherwise create a standard function
    node = asty.AsyncFunctionDef if is_async else asty.FunctionDef
    name = compiler.get_anon_var()
    ret += compile_function_node(
        compiler, expr, node, [], tp, name, args, returns, body, scope
    )

    # return its name as the final expr
    return ret + Result(expr=ret.temp_variables[0])


@pattern_macro("defn", [
    maybe(keepsym(":async")),
    maybe(brackets(many(FORM))),
    maybe(type_params),
    maybe_annotated(SYM),
    lambda_list,
    many(FORM)])
def compile_function_def(compiler, expr, root, is_async, decorators, tp, name, params, body):
    name, returns = name
    node = asty.AsyncFunctionDef if is_async else asty.FunctionDef
    decorators, ret, _ = compiler._compile_collect(decorators[0] if decorators else [])
    args, ret2 = compile_lambda_list(compiler, params)
    ret += ret2
    name = mangle(compiler._nonconst(name))
    compiler.scope.define(name)
    with compiler.local_state(), compiler.scope.create(ScopeFn, args, is_async) as scope:
        body = compiler._compile_branch(body)

    return ret + compile_function_node(
        compiler, expr, node, decorators, tp, name, args, returns, body, scope
    )


def compile_function_node(compiler, expr, node, decorators, tp, name, args, returns, body, scope):
    ret = Result()

    if body.expr:
        # implicitly return final expression,
        # except for async generators
        enode = asty.Expr if scope.is_async and scope.has_yield else asty.Return
        body += enode(body.expr, value=body.expr)

    ret += node(
        expr,
        name=name,
        args=args,
        body=body.stmts or [asty.Pass(expr)],
        decorator_list=decorators,
        returns=compiler.compile(returns).force_expr if returns is not None else None,
        **digest_type_params(compiler, tp),
    )

    ast_name = asty.Name(expr, id=name, ctx=ast.Load())
    return ret + Result(temp_variables=[ast_name, ret.stmts[-1]])


@pattern_macro(
    "defmacro",
    [
        SYM,
        brackets(
            maybe(many(argument) + sym("/")),
            many(argument),
            maybe(varargs("unpack-iterable", NASYM)),
        ),
        many(FORM),
    ],
)
def compile_macro_def(compiler, expr, root, name, params, body):
    def E(*x): return Expression(x)
    S = Symbol

    compiler.warn_on_core_shadow(name)
    fn_def = E(S("fn"), List(expr[2]), *body).replace(expr)
    if compiler.is_in_local_state():
        # We're in a local scope, so define the new macro locally.
        state = compiler.local_state_stack[-1]
        # Produce code that will set the macro to a local variable.
        ret = compiler.compile(E(
            S("setv"),
            S(local_macro_name(name)),
            fn_def).replace(expr))
        # Also evaluate the macro definition now, and put it in
        # state['macros'].
        state['macros'][mangle(name)] = compiler.eval(fn_def)
        return ret + ret.expr_as_stmt()
    # Otherwise, define the macro module-wide.
    ret = compiler.compile(E(S("eval-and-compile"), E(
        E(dotted("hy.macros.macro"), str(name)),
        fn_def)).replace(expr))
    return ret + ret.expr_as_stmt()


def compile_lambda_list(compiler, params):
    ret = Result()
    posonly_parms, args_parms, rest_parms, kwonly_parms, kwargs_parms = params

    if not (posonly_parms or posonly_parms is None):
        compiler._syntax_error(
            params, "at least one argument must precede /"
        )

    posonly_parms = posonly_parms or []

    is_positional_arg = lambda x: isinstance(x[0], Symbol)
    invalid_non_default = next(
        (
            arg
            for arg in dropwhile(is_positional_arg, posonly_parms + args_parms)
            if is_positional_arg(arg)
        ),
        None,
    )
    if invalid_non_default:
        compiler._syntax_error(
            invalid_non_default[0], "non-default argument follows default argument"
        )

    posonly_ast, posonly_defaults, ret = compile_arguments_set(
        compiler, posonly_parms, ret
    )
    args_ast, args_defaults, ret = compile_arguments_set(compiler, args_parms, ret)
    kwonly_ast, kwonly_defaults, ret = compile_arguments_set(
        compiler, kwonly_parms, ret, True
    )
    rest_ast = kwargs_ast = None

    if rest_parms == Symbol("*"):  # rest is a positional only marker
        if not kwonly_parms:
            compiler._syntax_error(
                rest_parms, "named arguments must follow bare *"
            )
        rest_ast = None
    elif rest_parms:  # rest is capturing varargs
        [rest_ast], _, ret = compile_arguments_set(compiler, [rest_parms], ret)
    if kwargs_parms:
        [kwargs_ast], _, ret = compile_arguments_set(compiler, [kwargs_parms], ret)

    return (
        ast.arguments(
            args=args_ast,
            defaults=[*posonly_defaults, *args_defaults],
            vararg=rest_ast,
            posonlyargs=posonly_ast,
            kwonlyargs=kwonly_ast,
            kw_defaults=kwonly_defaults,
            kwarg=kwargs_ast,
        ),
        ret,
    )


def compile_arguments_set(compiler, decls, ret, is_kwonly=False):
    args_ast = []
    args_defaults = []

    for decl, ann in decls:
        default = None

        # funcparserlib will check to make sure that the only times we
        # ever have a hy List here are due to a default value.
        if isinstance(decl, List):
            sym, default = decl
        else:
            sym = decl

        if ann is not None:
            ret += compiler.compile(ann)
            ann_ast = ret.force_expr
        else:
            ann_ast = None

        if default is not None:
            ret += compiler.compile(default)
            args_defaults.append(ret.force_expr)
        # Kwonly args without defaults are considered required
        elif not isinstance(decl, List) and is_kwonly:
            args_defaults.append(None)
        elif isinstance(decl, List):
            # Note that the only time any None should ever appear here
            # is in kwargs, since the order of those with defaults vs
            # those without isn't significant in the same way as
            # positional args.
            args_defaults.append(None)

        args_ast.append(
            asty.arg(sym, arg=mangle(compiler._nonconst(sym)), annotation=ann_ast)
        )

    return args_ast, args_defaults, ret


@pattern_macro("return", [maybe(FORM)])
def compile_return(compiler, expr, root, arg):
    ret = Result()
    if arg is None:
        return asty.Return(expr, value=None)
    ret += compiler.compile(arg)
    return ret + asty.Return(expr, value=ret.force_expr)


@pattern_macro("yield", [times(0, 2, FORM)])
def compile_yield_expression(compiler, expr, root, args):
    if is_inside_function_scope(compiler.scope):
        nearest_python_scope(compiler.scope).has_yield = True
    yield_from = False
    if len(args) == 2:
        from_kw, x = args
        if from_kw != Keyword("from"):
            compiler._syntax_error(from_kw, "two-argument `yield` requires `:from`")
        yield_from = True
        args = [x]
    ret = Result()
    if args:
        ret += compiler.compile(args[0])
    return ret + (asty.YieldFrom if yield_from else asty.Yield)(expr, value=ret.force_expr)


@pattern_macro("await", [FORM])
def compile_yield_from_or_await_expression(compiler, expr, root, arg):
    ret = Result() + compiler.compile(arg)
    return ret + asty.Await(expr, value=ret.force_expr)


# ------------------------------------------------
# * `defclass`
# ------------------------------------------------


@pattern_macro(
    "defclass",
    [
        maybe(brackets(many(FORM))),
        maybe(type_params),
        SYM,
        maybe(brackets(many(FORM)) + maybe(STR) + many(FORM)),
    ],
)
def compile_class_expression(compiler, expr, root, decorators, tp, name, rest):
    base_list, docstring, body = rest or ([[]], None, [])

    decorators, ret, _ = compiler._compile_collect(decorators[0] if decorators else [])
    bases_expr, ret2, keywords = compiler._compile_collect(
        base_list[0], with_kwargs=True
    )
    ret += ret2

    bodyr = Result()

    if docstring is not None:
        bodyr += compiler.compile(docstring).expr_as_stmt()

    name = mangle(compiler._nonconst(name))
    compiler.scope.define(name)

    with compiler.local_state(), compiler.scope.create(ScopeFn):
        e = compiler._compile_branch(body)
        bodyr += e + e.expr_as_stmt()

    return ret + asty.ClassDef(
        expr,
        decorator_list=decorators,
        name=name,
        keywords=keywords,
        bases=bases_expr,
        body=bodyr.stmts or [asty.Pass(expr)],
        **digest_type_params(compiler, tp)
    )


# ------------------------------------------------
# * `import` and `require`
# ------------------------------------------------

module_name_pattern = SYM | pexpr(
    some(lambda x: isinstance(x, Symbol) and not str(x[0]).strip(".")), oneplus(SYM)
)


def module_name_str(x):
    return (
        ".".join(map(mangle, x[1][x[1][0] == Symbol("None") :]))
        if isinstance(x, Expression)
        else str(x)
        if isinstance(x, Symbol) and not x.strip(".")
        else mangle(x)
    )


importlike = (
    keepsym("*")
    | (keepsym(":as") + SYM)
    | brackets(many(SYM + maybe(sym(":as") + SYM)))
)


def assignment_shape(module, rest):
    prefix = ""
    assignments = "EXPORTS"
    if rest is None:
        # (import foo)
        prefix = module_name_str(module)
    elif rest == Symbol("*"):
        # (import foo *)
        pass
    elif rest[0] == Keyword("as"):
        # (import foo :as bar)
        prefix = mangle(rest[1])
    else:
        # (import foo [bar baz :as MyBaz bing])
        assignments = [(k, v or k) for k, v in rest[0]]
    return prefix, assignments


@pattern_macro(
    "require",
    [
        many(
            module_name_pattern
            + times(
                0,
                2,
                (maybe(sym(":macros")) + importlike)
                | (keepsym(":readers") + (keepsym("*") | brackets(many(SYM)))),
            )
        )
    ],
)
def compile_require(compiler, expr, root, entries):
    ret = Result()
    for entry in entries:
        module, assignments = entry
        readers, rest = (
            [names for key, names in assignments if (key == Keyword("readers")) == flag]
            for flag in (True, False)
        )
        if len(rest) > 1 or len(readers) > 1:
            compiler._syntax_error(
                entry,
                f"redefinition of ':{'macros' if len(rest) > 1 else 'readers'}' brackets.",
            )

        rest = rest[0] if rest else None  # None used for prefixed macro import
        readers = readers and readers[0]

        prefix, assignments = assignment_shape(module, rest)
        module_name = module_name_str(module)
        if isinstance(module, Expression) and module[1][0] == Symbol("None"):
            # Prepend leading dots to `module_name`.
            module_name = str(module[0]) + module_name

        # we don't want to import all macros as prefixed if we're specifically
        # importing readers but not macros
        # (require a-module :readers ["!"])
        if (rest or not readers) and compiler.is_in_local_state():
            reqs = require(
                module_name,
                compiler.local_state_stack[-1]['macros'],
                assignments = assignments,
                prefix = prefix,
                compiler = compiler)
            ret += compiler.compile(Expression([
                Symbol("setv"),
                List([Symbol(local_macro_name(m)) for m, _, _ in reqs]),
                Expression([
                    dotted("hy.macros.require_vals"),
                    String(module_name),
                    Dict(),
                    Keyword("assignments"),
                    List([(String(m), String(m)) for _, m, _ in reqs])])]).replace(expr))
            ret += ret.expr_as_stmt()
        elif (rest or not readers) and require(
                module_name,
                compiler.module,
                assignments = assignments,
                prefix = prefix,
                compiler = compiler):
            # Actually calling `require` is necessary for macro expansions
            # occurring during compilation.
            # The `require` we're creating in AST is the same as above, but used at
            # run-time (e.g. when modules are loaded via bytecode).
            ret += compiler.compile(
                Expression(
                    [
                        dotted("hy.macros.require"),
                        String(module_name),
                        Symbol("None"),
                        Keyword("target_module_name"),
                        String(compiler.module.__name__),
                        Keyword("assignments"),
                        (
                            String("EXPORTS")
                            if assignments == "EXPORTS"
                            else List([List([String(k), String(v)]) for k, v in assignments])
                        ),
                        Keyword("prefix"),
                        String(prefix),
                    ]
                ).replace(expr)
            )
            ret += ret.expr_as_stmt()

        if readers:
            reader_assignments = (
                "ALL"
                if readers == Symbol("*")
                else [str(reader) for reader in readers[0]]
            )
            if require_reader(module_name, compiler.module, reader_assignments):
                ret += compiler.compile(
                    mkexpr(
                        "do",
                        mkexpr(
                            dotted("hy.macros.require-reader"),
                            String(module_name),
                            "None",
                            [reader_assignments],
                        ),
                        mkexpr(
                            "eval-when-compile",
                            mkexpr(
                                dotted("hy.macros.enable-readers"),
                                "None",
                                mkexpr(dotted("hy.reader.HyReader.current-reader")),
                                [reader_assignments],
                            ),
                        ),
                    ).replace(expr)
                )
                ret += ret.expr_as_stmt()
    return ret


@pattern_macro("import", [maybe(keepsym(":lazy")), many(module_name_pattern + maybe(importlike))])
def compile_import(compiler, expr, root, is_lazy, entries):

    if not PY3_15 and is_lazy:
        compiler._syntax_error(is_lazy, "Lazy imports require Python 3.15 or later")
    is_lazy = dict(is_lazy = True) if is_lazy else {}

    ret = Result()

    for entry in entries:
        module, _ = entry
        prefix, assignments = assignment_shape(*entry)
        module_name = module_name_str(module)
        if assignments == "EXPORTS" and prefix == "":
            node = asty.ImportFrom
            names = [asty.alias(module, name="*", asname=None)]
        elif assignments == "EXPORTS":
            compiler.scope.define(prefix)
            node = asty.Import
            names = [asty.alias(
                module,
                name = module_name,
                asname = prefix if prefix != module_name else None)]
        else:
            node = asty.ImportFrom
            names = []
            for k, v in assignments:
                compiler.scope.define(mangle(v))
                names.append(asty.alias(
                    module,
                    name = mangle(k),
                    asname = None if v == k else mangle(v)))
        ret += node(
            expr,
            names = names,
            **is_lazy,
            **({} if node is asty.Import else dict(
                module = module_name
                    if module_name and module_name.strip(".")
                    else None,
                level =
                    len(module[0])
                    if isinstance(module, Expression)
                        and module[1][0] == Symbol("None")
                    else len(module)
                    if isinstance(module, Symbol)
                        and not module.strip(".")
                    else 0)))

    return ret


# ------------------------------------------------
# * Miscellany
# ------------------------------------------------


@pattern_macro("assert", [FORM, maybe(FORM)])
def compile_assert_expression(compiler, expr, root, test, msg):
    test = compiler.compile(test)
    if msg:
        msg = compiler.compile(msg)

    if not (test.stmts or (msg and msg.stmts)):
        return asty.Assert(expr, test=test.force_expr, msg=msg and msg.force_expr)

    return asty.If(
        expr,
        test=asty.Name(expr, id="__debug__", ctx=ast.Load()),
        orelse=[],
        body=test.stmts
        + [
            asty.If(
                test,
                test=asty.UnaryOp(test, op=ast.Not(), operand=test.force_expr),
                orelse=[],
                body=(msg.stmts if msg else [])
                + [
                    asty.Assert(
                        expr,
                        test=asty.Constant(test, value=False),
                        msg=msg and msg.force_expr,
                    )
                ],
            )
        ],
    )


@pattern_macro(
    "unquote unquote-splice unpack-mapping except except* finally else".split(),
    [many(FORM)],
)
def compile_placeholder(compiler, expr, root, body):
    raise ValueError(f"`{root}` is not allowed here")



"Tooling for reading/parsing source character-by-character."

import itertools
import re
from collections import deque
from contextlib import contextmanager

from .exceptions import PrematureEndOfInput

_whitespace = re.compile(r"[ \t\n\r\f\v]+")


def isnormalizedspace(s):
    return bool(_whitespace.match(s))


class ReaderMeta(type):
    """Provides a class with a dispatch map `DEFAULT_TABLE`
    and a decorator `@reader_for`."""

    @classmethod
    def __prepare__(cls, name, bases):
        namespace = super().__prepare__(cls, name, bases)
        namespace["reader_for"] = cls._attach_reader
        return namespace

    @staticmethod
    def _attach_reader(char, args=None):
        def wrapper(f):
            handler = f if args is None else f(*args)
            f._readers = {**getattr(f, "_readers", {}), char: handler}
            return f

        return wrapper

    def __new__(cls, name, bases, namespace):
        del namespace["reader_for"]
        default_table = {}
        for base in bases:
            if hasattr(base, "DEFAULT_TABLE"):
                default_table |= base.DEFAULT_TABLE
        for method in namespace.values():
            if callable(method) and hasattr(method, "_readers"):
                default_table.update(method._readers)
        namespace["DEFAULT_TABLE"] = default_table
        return super().__new__(cls, name, bases, namespace)


class Reader(metaclass=ReaderMeta):
    """An abstract base class for reading input character-by-character.

    See :py:class:`hy.HyReader` for an example
    of creating a reader class.

    Attributes:
        ends_ident (set[str]):
            The set of characters that indicate the end of an identifier
        reader_table (dict[str, Callable]):
            A dictionary mapping a reader-macro key to its dispatch function
        pos (tuple[int, int]):
            A read-only (line, column) tuple indicating the current cursor
            position of the source being read"""

    __module__ = 'hy'

    def __init__(self):
        self._source = None
        self._filename = None

        self.ends_ident = set(self.NON_IDENT)
        self.reader_table = self.DEFAULT_TABLE.copy()

    def _set_source(self, stream=None, filename=None):
        if filename is not None:
            self._filename = filename
        if stream is not None:
            pos = stream.tell()
            self._source = stream.read()
            stream.seek(pos)

            self._stream = stream
            self._peek_chars = deque()
            self._saved_chars = []
            self._pos = (1, 0)
            self._eof_tracker = self._pos

    @property
    def pos(self):
        return self._pos

    @contextmanager
    def end_identifier(self, character):
        """A context manager to temporarily add a new character to the
        :py:attr:`ends_ident` set."""

        prev_ends_ident = self.ends_ident.copy()
        self.ends_ident.add(character)
        try:
            yield
        finally:
            self.ends_ident = prev_ends_ident

    ###
    # Character streaming
    ###

    @contextmanager
    def saving_chars(self):
        """A context manager to save all read characters. The value is a list
        of characters, rather than a single string."""

        self._saved_chars.append([])
        yield self._saved_chars[-1]
        saved = self._saved_chars.pop()
        if self._saved_chars:
            # `saving_chars` is being used recursively. The
            # characters we collected for the inner case should also
            # be saved for the outer case.
            self._saved_chars[-1].extend(saved)

    def peekc(self):
        "Peek at the next character, returning it but not consuming it."
        if self._peek_chars:
            return self._peek_chars[-1]
        nc = self._stream.read(1)
        self._peek_chars.append(nc)
        return nc

    def peeking(self, eof_ok=False):
        """As :func:`chars`, but without consuming any of the returned
        characters. This method is useful for looking several characters
        ahead."""

        for nc in reversed(self._peek_chars):
            yield nc
        while True:
            c = self._stream.read(1)
            if not c:
                break
            self._peek_chars.appendleft(c)
            yield c
        if not c and not eof_ok:
            raise PrematureEndOfInput.from_reader(
                "Premature end of input while peeking", self
            )

    def getc(self):
        """Consume one character from the stream and return it. This method
        does the bookkeeping for position data, so all character consumption
        should go through it."""

        c = self.peekc()
        self._peek_chars.pop()

        if c:
            line, col = self._pos
            col += 1
            if c == "\n":
                line += 1
                col = 0
            self._pos = (line, col)

            if not isnormalizedspace(c):
                self._eof_tracker = self._pos

        if self._saved_chars:
            self._saved_chars[-1].append(c)

        return c

    def peek_and_getc(self, target):
        """Peek at the next character and check if it's equal to ``target``,
        only consuming it if it's equal. A :py:class:`bool` is returned."""

        nc = self.peekc()
        if nc == target:
            self.getc()
            return True
        return False

    def chars(self, eof_ok=False):
        """Consume and yield characters of the stream. If ``eof_ok``
        is false (the default) and the end of the stream is reached,
        raise :exc:`hy.PrematureEndOfInput`."""

        while True:
            c = self.getc()
            if not c:
                break
            yield c
        if not c and not eof_ok:
            raise PrematureEndOfInput.from_reader(
                "Premature end of input while streaming chars", self
            )

    ###
    # Reading multiple characters
    ###

    def getn(self, n):
        "Consume and return ``n`` characters."
        return "".join(itertools.islice(self.chars(), n))

    def slurp_space(self):
        "Consume and return zero or more whitespace characters."
        n = 0
        for c in self.peeking(eof_ok=True):
            if not isnormalizedspace(c):
                break
            n += 1
        return self.getn(n)

    def read_ident(self, just_peeking=False):
        """Read characters until we hit something in :py:attr:`ends_ident`. The
        characters are consumed unless ``just_peeking`` is true."""

        ident = []
        for nc in self.peeking(eof_ok=True):
            if not nc or nc in self.ends_ident or isnormalizedspace(nc):
                # `not nc` means EOF, but that's okay.
                break
            ident.append(nc)
        if not just_peeking:
            self.getn(len(ident))
        return "".join(ident)

    ###
    # Reader dispatch logic
    ###

    def dispatch(self, tag):
        """Call the handler for the reader macro with key ``tag`` (a
        string). Return the model it produces, if any."""

        return self.reader_table[tag](self, tag)



import ast
import builtins
import code
import codeop
import hashlib
import importlib
import linecache
import os
import platform
import sys
import time
import traceback
import types
from contextlib import contextmanager

import hy
from hy.compat import PY3_12
from hy.compiler import HyASTCompiler, hy_compile
from hy.completer import Completer, completion
from hy.errors import (
    HyLanguageError,
    HyMacroExpansionError,
    HyRequireError,
    filtered_hy_exceptions,
)
from hy.importer import HyLoader
from hy.macros import enable_readers, require, require_reader
from hy.reader import mangle, read_many
from hy.reader.exceptions import PrematureEndOfInput
from hy.reader.hy_reader import HyReader


class HyQuitter:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Use (%s) or Ctrl-D (i.e. EOF) to exit" % (self.name)

    __str__ = __repr__

    def __call__(self, code=None):
        try:
            sys.stdin.close()
        except:
            pass
        raise SystemExit(code)


class HyHelper:
    def __repr__(self):
        return (
            "Use (help) for interactive help, or (help object) for help "
            "about object."
        )

    def __call__(self, *args, **kwds):
        import pydoc

        return pydoc.help(*args, **kwds)


@contextmanager
def extend_linecache(add_cmdline_cache):
    _linecache_checkcache = linecache.checkcache

    def _cmdline_checkcache(*args):
        _linecache_checkcache(*args)
        linecache.cache.update(add_cmdline_cache)

    linecache.checkcache = _cmdline_checkcache
    yield
    linecache.checkcache = _linecache_checkcache


_codeop_maybe_compile = codeop._maybe_compile
codeop._maybe_compile = (lambda compiler, source, filename, symbol, *args, **kwargs:
    # Python 3.14 adds a mandatory parameter `flags`, which is
    # sometimes specified by position and sometimes by name.
    compiler(source, filename, symbol)
    if isinstance(compiler, HyCompile) else
    _codeop_maybe_compile(compiler, source, filename, symbol, *args, **kwargs))


class HyCompile(codeop.Compile):
    """This compiler uses `linecache` like
    `IPython.core.compilerop.CachingCompiler`.
    """

    def __init__(
        self, module, locals, ast_callback=None, hy_compiler=None, cmdline_cache={}
    ):
        self.module = module
        self.locals = locals
        self.ast_callback = ast_callback
        self.hy_compiler = hy_compiler
        self.reader = HyReader()
        self.skip_next_shebang = False

        super().__init__()

        if hasattr(self.module, "_hy_reader_macros"):
            enable_readers(
                self.module, self.reader, self.module._hy_reader_macros.keys()
            )

        self.cmdline_cache = cmdline_cache

    def _cache(self, source, name):
        entry = (
            len(source),
            time.time(),
            [line + "\n" for line in source.splitlines()],
            name,
        )

        linecache.cache[name] = entry
        self.cmdline_cache[name] = entry

    def _update_exc_info(self):
        t, v, tb = set_last_exc()
        # Skip our frame.
        self.locals["_hy_exc_info"] = (t, v, getattr(tb, "tb_next", tb))

    def __call__(self, source, filename="<input>", symbol=None):
        symbol = "exec"
          # This parameter is required by `codeop.Compile`, but we
          # ignore it in favor of always using "exec".

        hash_digest = hashlib.sha1(source.encode("utf-8").strip()).hexdigest()
        name = "{}-{}".format(filename.strip("<>"), hash_digest)

        self._cache(source, name)

        try:
            # Our compiler doesn't correspond to a real, fixed source file, so
            # we need to [re]set these.
            self.hy_compiler.filename = name
            self.hy_compiler.source = source
            hy_ast = read_many(
                source, filename=name, reader=self.reader,
                skip_shebang=self.skip_next_shebang,
            )
            self.skip_next_shebang = False
            exec_ast, eval_ast = hy_compile(
                hy_ast,
                self.module,
                root=ast.Module,
                get_expr=True,
                compiler=self.hy_compiler,
                filename=name,
                source=source,
                import_stdlib=False,
            )

            if self.ast_callback:
                self.ast_callback(exec_ast, eval_ast)

            exec_code = super().__call__(exec_ast, name, symbol)
            eval_code = super().__call__(eval_ast, name, "eval")

        except Exception as e:
            self._update_exc_info()
            if isinstance(e, (PrematureEndOfInput, SyntaxError)):
                raise
            else:
                # Hy will raise exceptions during compile-time that Python would
                # raise during run-time (e.g. import errors for `require`).  In
                # order to work gracefully with the Python world, we convert such
                # Hy errors to code that purposefully reraises those exceptions in
                # the places where Python code expects them.
                # Capture a traceback without the compiler/REPL frames.
                exec_code = super(HyCompile, self).__call__(
                    "raise _hy_exc_info[1].with_traceback(_hy_exc_info[2])",
                    name,
                    symbol,
                )
                eval_code = super(HyCompile, self).__call__("None", name, "eval")

        return exec_code, eval_code


class HyCommandCompiler(codeop.CommandCompiler):
    def __init__(self, *args, allow_incomplete = True, **kwargs):
        self.compiler = HyCompile(*args, **kwargs)
        self.allow_incomplete = allow_incomplete

    def __call__(self, *args, **kwargs):
        try:
            return super().__call__(*args, **kwargs)
        except PrematureEndOfInput:
            # We have to do this here, because `codeop._maybe_compile` won't
            # take `None` for a return value (at least not in Python 2.7) and
            # this exception type is also a `SyntaxError`, so it will be caught
            # by `code.InteractiveConsole` base methods before it reaches our
            # `runsource`.
            if not self.allow_incomplete:
                raise


class REPL(code.InteractiveConsole):
    """A subclass of :class:`code.InteractiveConsole` for Hy.

    A convenient way to use this class to interactively debug code is to insert the
    following in the code you want to debug::

        (.run (hy.REPL :locals {#** (globals) #** (locals)}))

    Or in Python:

    .. code-block:: python

       import hy; hy.REPL(locals = {**globals(), **locals()}).run()

    Note that as with :func:`code.interact`, changes to local variables inside the
    REPL are not propagated back to the original scope."""

    __module__ = 'hy'

    def __init__(self, spy=False, spy_delimiter=('-' * 30), output_fn=None, locals=None, filename="<stdin>", allow_incomplete=True):

        # Create a proper module for this REPL so that we can obtain it easily
        # (e.g. using `importlib.import_module`).
        # We let `InteractiveConsole` initialize `self.locals` when it's
        # `None`.
        super().__init__(locals=locals, filename=filename)

        module_name = self.locals.get("__name__", "__console__")
        # Make sure our newly created module is properly introduced to
        # `sys.modules`, and consistently use its namespace as `self.locals`
        # from here on.
        self.module = sys.modules.setdefault(module_name, types.ModuleType(module_name))
        self.module.__dict__.update(self.locals)
        self.locals = self.module.__dict__

        self.ps1 = "=> "
        self.ps2 = "... "

        if os.environ.get("HYSTARTUP"):
            try:
                loader = HyLoader("__hystartup__", os.environ.get("HYSTARTUP"))
                spec = importlib.util.spec_from_loader(loader.name, loader)
                mod = importlib.util.module_from_spec(spec)
                sys.modules.setdefault(mod.__name__, mod)
                loader.exec_module(mod)
                imports = mod.__dict__.get(
                    "__all__",
                    [name for name in mod.__dict__ if not name.startswith("_")],
                )
                imports = {name: mod.__dict__[name] for name in imports}
                spy = spy or imports.get("repl_spy")
                output_fn = output_fn or imports.get("repl_output_fn")
                self.ps1 = imports.get("repl_ps1", self.ps1)
                self.ps2 = imports.get("repl_ps2", self.ps2)

                # Load imports and defs
                self.locals.update(imports)

                # load module macros
                require(mod, self.module, assignments="ALL")
                require_reader(mod, self.module, assignments="ALL")
            except Exception as e:
                print(e)

        self.hy_compiler = HyASTCompiler(self.module, module_name)

        self.cmdline_cache = {}
        self.compile = HyCommandCompiler(
            self.module,
            self.locals,
            ast_callback=self.ast_callback,
            hy_compiler=self.hy_compiler,
            cmdline_cache=self.cmdline_cache,
            allow_incomplete=allow_incomplete,
        )

        self.spy = spy
        self.spy_delimiter = spy_delimiter
        self.last_value = None
        self.print_last_value = True

        if output_fn is None:
            self.output_fn = hy.repr
        elif callable(output_fn):
            self.output_fn = output_fn
        elif "." in output_fn:
            parts = [mangle(x) for x in output_fn.split(".")]
            module, f = ".".join(parts[:-1]), parts[-1]
            self.output_fn = getattr(importlib.import_module(module), f)
        else:
            self.output_fn = getattr(builtins, mangle(output_fn))

        # Pre-mangle symbols for repl recent results: *1, *2, *3
        self._repl_results_symbols = [mangle("*{}".format(i + 1)) for i in range(3)]
        self.locals.update({sym: None for sym in self._repl_results_symbols})

        # Allow access to the running REPL instance
        self.locals["_hy_repl"] = self

        # Compile an empty statement to load the standard prelude
        exec_ast = hy_compile(
            read_many(""), self.module, compiler=self.hy_compiler, import_stdlib=True
        )
        if self.ast_callback:
            self.ast_callback(exec_ast, None)

    def ast_callback(self, exec_ast, eval_ast):
        if self.spy:
            try:
                # Mush the two AST chunks into a single module for
                # conversion into Python.
                new_ast = ast.Module(
                    exec_ast.body
                    + ([] if eval_ast is None else [ast.Expr(eval_ast.body)]),
                    type_ignores=[],
                )
                print(ast.unparse(new_ast))
                print(self.spy_delimiter)
            except Exception:
                msg = "Exception in AST callback:\n{}\n".format(traceback.format_exc())
                self.write(msg)

    def _error_wrap(self, exc_info_override=False, *args, **kwargs):
        # When `exc_info_override` is true, use a traceback that
        # doesn't have the REPL frames.
        t, v, tb = set_last_exc(
            exc_info_override and self.locals.get("_hy_exc_info"))
        sys.excepthook(t, v, tb)
        self.locals[mangle("*e")] = v

    def showsyntaxerror(self, filename=None, source=None):
        if filename is None:
            filename = self.filename
        self.print_last_value = False

        self._error_wrap(exc_info_override=True, filename=filename)

    def showtraceback(self):
        self._error_wrap()

    def runcode(self, code):
        try:
            eval(code[0], self.locals)
            self.last_value = eval(code[1], self.locals)
            # Don't print `None` values.
            self.print_last_value = self.last_value is not None
        except SystemExit:
            raise
        except Exception as e:
            # Set this to avoid a print-out of the last value on errors.
            self.print_last_value = False
            self.showtraceback()

    def runsource(self, source, filename="<stdin>", symbol="exec"):
        try:
            res = super().runsource(source, filename, symbol)
        except (HyMacroExpansionError, HyRequireError):
            # We need to handle these exceptions ourselves, because the base
            # method only handles `OverflowError`, `SyntaxError` and
            # `ValueError`.
            self.showsyntaxerror(filename)
            return False
        except (HyLanguageError):
            # Our compiler will also raise `TypeError`s
            self.showtraceback()
            return False

        # Shift exisitng REPL results
        if not res:
            next_result = self.last_value
            for sym in self._repl_results_symbols:
                self.locals[sym], next_result = next_result, self.locals[sym]

            # Print the value.
            if self.print_last_value:
                try:
                    output = self.output_fn(self.last_value)
                except Exception:
                    self.showtraceback()
                    return False

                print(output)

        return res

    def run(self):
        "Start running the REPL. Return 0 when done."

        sentinel = []
        saved_values = (
            getattr(sys, "ps1", sentinel),
            getattr(sys, "ps2", sentinel),
            builtins.quit,
            builtins.exit,
            builtins.help,
        )
        try:
            sys.ps1 = self.ps1
            sys.ps2 = self.ps2
            builtins.quit = HyQuitter("quit")
            builtins.exit = HyQuitter("exit")
            builtins.help = HyHelper()

            namespace = self.locals
            with filtered_hy_exceptions(), extend_linecache(
                self.cmdline_cache
            ), completion(Completer(namespace)):
                self.interact(self.banner())

        finally:
            sys.ps1, sys.ps2, builtins.quit, builtins.exit, builtins.help = saved_values
            for a in "ps1", "ps2":
                if getattr(sys, a) is sentinel:
                    delattr(sys, a)

        return 0

    def banner(self):
        return "Hy {version}{nickname} using {py}({build}) {pyversion} on {os}".format(
            version=hy.__version__,
            nickname="" if hy.nickname is None else f' ({hy.nickname})',
            py=platform.python_implementation(),
            build=platform.python_build()[0],
            pyversion=platform.python_version(),
            os=platform.system(),
        )


def set_last_exc(exc_info = None):
    """Setting `sys.last_exc`, or `sys.last_type` on earlier Pythons,
    makes it easier for the user to call the debugger."""
    t, v, tb = exc_info or sys.exc_info()
    if PY3_12:
          sys.last_exc = v
    else:
          sys.last_type, sys.last_value, sys.last_traceback = t, v, tb
    return t, v, tb
